# -*- coding: utf-8 -*-
#
#   Paraqus - A VTK exporter for FEM results.
#
#   Copyright (C) 2022, Furlan, Stollberg and Menzel
#
#    This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along with this program. If not, see https://www.gnu.org/licenses/.
"""
Collection of constants defined for Paraqus.

The constants are derived from a custom type to make comparisons easier, e.g.
when user input is compared to constants.

"""

import sys

class ParaqusConstant(object):
    """
    A type for named constants.

    Parameters
    ----------
    name : str
        Name of the constant, used for comparisons.

    Attributes
    ----------
    name : str
        Name of the constant, used for comparisons.

    """
    def __init__(self, name):
        self.name = name

    def __str__(self):
        return str(self.name)

    def __eq__(self, other):
        if not isinstance(other, ParaqusConstant):
            return str(self.name).upper() == str(other).upper()
        else:
            return self.name == other.name

    def __add__(self, other):
        return str(self.name) + str(other)

    def __radd__(self, other):
        return str(other) + str(self.name)


# Define constants
UINT64 = ParaqusConstant("UINT64")
UINT32 = ParaqusConstant("UINT32")

BINARY = ParaqusConstant("BINARY")
ASCII = ParaqusConstant("ASCII")
BASE64 = ParaqusConstant("BASE64")
RAW = ParaqusConstant("RAW")
LITTLE_ENDIAN = ParaqusConstant("LittleEndian")
BIG_ENDIAN = ParaqusConstant("BigEndian")
BYTE_ORDER = LITTLE_ENDIAN if sys.byteorder == "little" else BIG_ENDIAN
BYTE_ORDER_CHAR = (ParaqusConstant("<") if BYTE_ORDER == LITTLE_ENDIAN
                   else ParaqusConstant(">"))

ELEMENTS = ParaqusConstant("ELEMENTS")
NODES = ParaqusConstant("NODES")
WHOLE_ELEMENTS = ParaqusConstant("WHOLE_ELEMENTS")

USER = ParaqusConstant("USER")
ABAQUS = ParaqusConstant("ABAQUS")

SCALAR = ParaqusConstant("SCALAR")
VECTOR = ParaqusConstant("VECTOR")
TENSOR = ParaqusConstant("TENSOR")

MEAN = ParaqusConstant("MEAN")
ABSMAX = ParaqusConstant("ABSMAX")



# -*- coding: utf-8 -*-
#
#   Paraqus - A VTK exporter for FEM results.
#
#   Copyright (C) 2022, Furlan, Stollberg and Menzel
#
#    This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along with this program. If not, see https://www.gnu.org/licenses/.
"""
Abaqus element library for paraqus.

The module constant `ABQ_ELEMENT_LIBRARY` maps the abaqus element names to
integer vtk element types.

"""

# Identifiers:
# 3: First-order line
# 4: Second-order line
# 5: First-order tri
# 9: First-order quad
# 10: First-order tet
# 12: First-order hex
# 13: First-order prism
# 14: First-order pyramid
# 22: Second-order tri
# 23: Second-order quad
# 24: Second-order tet
# 25: Second-order hex
# 26: Second-order prism
# 28: Bi-quadratic quad

# We copied all the elements from the abaqus element library here. Note that
# not all elements are actually tested.

ABQ_ELEMENT_LIBRARY = {

# =============================================================================
#
#
#                   CONTINUUM ELEMENTS
#
#
# =============================================================================


#%% 1d solid elements
# -----------------------------------------------------------------------------

    # Diffusive heat transfer elements
    "DC1D2": 3,
    "DC1D3": 4,

    # Forced convection heat transfer elements
    "DCC1D2": 3,
    "DCC1D2D": 3,

    # Coupled thermal-electrical elements
    "DC1D2E": 3,
    "DC1D3E": 3,

    # Acoustic elements
    "AC1D2": 3,
    "AC1D3": 3,


#%% 2d solid elements
# -----------------------------------------------------------------------------

    # Plane strain elements
    "CPE3": 5,
    "CPE3H": 5,
    "CPE4": 9,
    "CPE4H": 9,
    "CPE4I": 9,
    "CPE4IH": 9,
    "CPE4R": 9,
    "CPE4RH": 9,
    "CPE6": 22,
    "CPE6H": 22,
    "CPE6M": 22,
    "CPE6MH": 22,
    "CPE8": 23,
    "CPE8H": 23,
    "CPE8R": 23,
    "CPE8RH": 23,

    # Plane stress elements
    "CPS3": 5,
    "CPS4": 9,
    "CPS4I": 9,
    "CPS4R": 9,
    "CPS6": 22,
    "CPS6M": 22,
    "CPS8": 23,
    "CPS8R": 23,

    # Generalized plane strain elements
    "CPEG3": 5,
    "CPEG3H": 5,
    "CPEG4": 9,
    "CPEG4H": 9,
    "CPEG4I": 9,
    "CPEG4IH": 9,
    "CPEG4R": 9,
    "CPEG4RH": 9,
    "CPEG6": 22,
    "CPEG6H": 22,
    "CPEG6M": 22,
    "CPEG6MH": 22,
    "CPEG8": 23,
    "CPEG8H": 23,
    "CPEG8R": 23,
    "CPEG8RH": 23,

    # Coupled temperature-displacement plane strain elements
    "CPE3T": 5,
    "CPE4T": 9,
    "CPE4HT": 9,
    "CPE4RT": 9,
    "CPE4RHT": 9,
    "CPE6MT": 22,
    "CPE6MHT": 22,
    "CPE8T": 23,
    "CPE8HT": 23,
    "CPE8RT": 23,
    "CPE8RHT": 23,

    # Coupled temperature-displacement plane stress elements
    "CPS3T": 5,
    "CPS4T": 9,
    "CPS4RT": 9,
    "CPS6MT": 22,
    "CPS8T": 23,
    "CPS8RT": 23,

    # Coupled temperature-displacement generalized plane strain elements
    "CPEG3T": 5,
    "CPEG3HT": 5,
    "CPEG4T": 9,
    "CPEG4HT": 9,
    "CPEG4RT": 9,
    "CPEG4RHT": 9,
    "CPEG6MT": 22,
    "CPEG6MHT": 22,
    "CPEG8T": 23,
    "CPEG8HT": 23,
    "CPEG8RHT": 23,

    # Diffusive heat transfer or mass diffusion elements
    "DC2D3": 5,
    "DC2D4": 9,
    "DC2D6": 22,
    "DC2D8": 23,

    # Forced convection/diffusion elements
    "DCC2D4": 5,
    "DCC2D4D": 5,

    # Coupled thermal-electrical elements
    "DC2D3E": 5,
    "DC2D4E": 9,
    "DC2D6E": 22,
    "DC2D8E": 23,

    # Pore pressure plane strain elements
    "CPE4P": 9,
    "CPE4PH": 9,
    "CPE4RP": 9,
    "CPE4RPH": 9,
    "CPE6MP": 22,
    "CPE6MPH": 22,
    "CPE8P": 23,
    "CPE8PH": 23,
    "CPE8RP": 23,
    "CPE8RPH": 23,

    # Coupled temperature–pore pressure plane strain elements
    "CPE4PT": 9,
    "CPE4PHT": 9,
    "CPE4RPT": 9,
    "CPE4RPHT": 9,

    # Acoustic elements
    "AC2D3": 5,
    "AC2D4": 9,
    "AC2D4R": 9,
    "AC2D6": 22,
    "AC2D8": 23,

    # Piezoelectric plane strain elements
    "CPE3E": 5,
    "CPE4E": 9,
    "CPE6E": 22,
    "CPE8E": 23,
    "CPE8RE": 23,

    # Piezoelectric plane stress elements
    "CPS3E": 5,
    "CPS4E": 9,
    "CPS6E": 22,
    "CPS8E": 23,
    "CPS8RE": 23,

    # Electromagnetic elements
    "EMC2D3": 5,
    "EMC2D4": 9,


#%% 3d solid elements
# -----------------------------------------------------------------------------

    # Stress/displacement elements
    "C3D4": 10,
    "C3D4H": 10,
    "C3D5": 14,
    "C3D5H": 14,
    "C3D6": 13,
    "C3D6H": 13,
    "C3D8": 12,
    "C3D8H": 12,
    "C3D8I": 12,
    "C3D8IH": 12,
    "C3D8R": 12,
    "C3D8RH": 12,
    "C3D8S": 12,
    "C3D8HS": 12,
    "C3D10": 24,
    "C3D10H": 24,
    "C3D10HS": 24,
    "C3D10M": 24,
    "C3D10MH": 24,
    "C3D15": 26,
    "C3D15H": 26,
    "C3D20": 25,
    "C3D20H": 25,
    "C3D20R": 25,
    "C3D20RH": 25,
    "CSS8": 12,

    # Coupled temperature-displacement elements
    "C3D4T": 10,
    "C3D6T": 13,
    "C3D6RT": 13,
    "C3D6HT": 13,
    "C3D6RHT": 13,
    "C3D8T": 12,
    "C3D8HT": 12,
    "C3D8RT": 12,
    "C3D8RHT": 12,
    "C3D10T": 24,
    "C3D10HT": 24,
    "C3D10MT": 24,
    "C3D10MHT": 24,
    "C3D20T": 25,
    "C3D20HT": 25,
    "C3D20RT": 25,
    "C3D20RHT": 25,

    # Coupled thermal-electrical-structural elements
    "Q3D4": 10,
    "Q3D6": 13,
    "Q3D8": 12,
    "Q3D8H": 12,
    "Q3D8R": 12,
    "Q3D8RH": 12,
    "Q3D10M": 24,
    "Q3D10MH": 24,
    "Q3D20": 25,
    "Q3D20H": 25,
    "Q3D20R": 25,
    "Q3D20RH": 25,

    # Diffusive heat transfer or mass diffusion elements
    "DC3D4": 10,
    "DC3D5": 14,
    "DC3D6": 13,
    "DC3D8": 12,
    "DC3D8R": 12,
    "DC3D10": 24,
    "DC3D15": 26,
    "DC3D20": 25,

    # Forced convection/diffusion elements
    "DCC3D8": 12,
    "DCC3D8D": 12,

    # Coupled thermal-electrical elements
    "DC3D4E": 10,
    "DC3D6E": 13,
    "DC3D8E": 12,
    "DC3D10E": 24,
    "DC3D15E": 26,
    "DC3D20E": 25,

    # Pore pressure elements
    "C3D4P": 10,
    "C3D4PH": 10,
    "C3D6P": 13,
    "C3D6PH": 13,
    "C3D8P": 12,
    "C3D8PH": 12,
    "C3D8RP": 12,
    "C3D8RPH": 12,
    "C3D10P": 24,
    "C3D10PH": 24,
    "C3D10MP": 24,
    "C3D10MPH": 24,
    "C3D20P": 25,
    "C3D20PH": 25,
    "C3D20RP": 25,
    "C3D20RPH": 25,

    # Coupled temperature–pore pressure elements
    "C3D4PT": 10,
    "C3D4PHT": 10,
    "C3D6PT": 13,
    "C3D8PT": 12,
    "C3D8PHT": 12,
    "C3D8RPT": 12,
    "C3D8RPHT": 12,
    "C3D10MPT": 24,
    "C3D10PT": 24,
    "C3D10PHT": 24,

    # Acoustic elements
    "AC3D4": 10,
    "AC3D5": 14,
    "AC3D6": 13,
    "AC3D8": 12,
    "AC3D8R": 12,
    "AC3D10": 24,
    "AC3D15": 26,
    "AC3D20": 25,

    # Poroelastic acoustic elements
    "C3D4A": 10,
    "C3D6A": 13,
    "C3D8A": 12,

    # Piezoelectric elements
    "C3D4E": 10,
    "C3D6E": 13,
    "C3D8E": 12,
    "C3D10E": 24,
    "C3D15E": 26,
    "C3D20E": 25,
    "C3D20RE": 25,

    # Electromagnetic elements
    "EMC3D4": 10,
    "EMC3D6": 13,
    "EMC3D8": 12,

    # Coupled Eulerian Lagrangian elements - mechanical only
    "EC3D8R": 12,

    # Coupled Eulerian Lagrangian elements - thermomechanical
    "EC3D8RT": 12,

#%% Axisymmetric solid elements

    # Stress/displacement elements without twist
    "CAX3": 5,
    "CAX3H": 5,
    "CAX4": 9,
    "CAX4H": 9,
    "CAX4I": 9,
    "CAX4IH": 9,
    "CAX4R": 9,
    "CAX4RH": 9,
    "CAX6": 22,
    "CAX6H": 22,
    "CAX6M": 22,
    "CAX6MH": 22,
    "CAX8": 23,
    "CAX8H": 23,
    "CAX8R": 23,
    "CAX8RH": 23,

    # Stress/displacement elements with twist
    "CGAX3": 5,
    "CGAX3H": 5,
    "CGAX4": 9,
    "CGAX4H": 9,
    "CGAX4R": 9,
    "CGAX4RH": 9,
    "CGAX6": 22,
    "CGAX6H": 22,
    "CGAX6M": 22,
    "CGAX6MH": 22,
    "CGAX8": 23,
    "CGAX8H": 23,
    "CGAX8R": 23,
    "CGAX8RH": 23,

    # Diffusive heat transfer or mass diffusion elements
    "DCAX3": 5,
    "DCAX4": 9,
    "DCAX6": 22,
    "DCAX8": 23,

    # Forced convection/diffusion elements
    "DCCAX2": 3,
    "DCCAX2D": 3,
    "DCCAX4": 9,
    "DCCAX4D": 9,

    # Coupled thermal-electrical elements
    "DCAX3E": 5,
    "DCAX4E": 9,
    "DCAX6E": 22,
    "DCAX8E": 23,

    # Coupled temperature-displacement elements without twist
    "CAX3T": 5,
    "CAX4T": 9,
    "CAX4HT": 9,
    "CAX4RT": 9,
    "CAX4RHT": 9,
    "CAX6MT": 22,
    "CAX6MHT": 22,
    "CAX8T": 23,
    "CAX8HT": 23,
    "CAX8RT": 23,
    "CAX8RHT": 23,

    # Coupled temperature-displacement elements with twist
    "CGAX3T": 5,
    "CGAX3HT": 5,
    "CGAX4T": 9,
    "CGAX4HT": 9,
    "CGAX4RT": 9,
    "CGAX4RHT": 9,
    "CGAX6MT": 22,
    "CGAX6MHT": 22,
    "CGAX8T": 23,
    "CGAX8HT": 23,
    "CGAX8RT": 23,
    "CGAX8RHT": 23,

    # Pore pressure elements
    "CAX4P": 9,
    "CAX4PH": 9,
    "CAX4RP": 9,
    "CAX4RPH": 9,
    "CAX6MP": 22,
    "CAX6MPH": 22,
    "CAX8P": 23,
    "CAX8PH": 23,
    "CAX8RP": 23,
    "CAX8RPH": 23,

    # Coupled temperature–pore pressure elements
    "CAX4PT": 9,
    "CAX4RPT": 9,
    "CAX4RPHT": 9,

    # Acoustic elements
    "ACAX3": 5,
    "ACAX4R": 9,
    "ACAX4": 9,
    "ACAX6": 22,
    "ACAX8": 23,

    # Piezoelectric elements
    "CAX3E": 5,
    "CAX4E": 9,
    "CAX6E": 22,
    "CAX8E": 23,
    "CAX8RE": 23,


# =============================================================================
#
#
#                   STRUCTURAL ELEMENTS
#
#
# =============================================================================


#%% Truss elements
# -----------------------------------------------------------------------------

    # 2D stress/displacement truss elements
    "T2D2": 3,
    "T2D2H": 3,
    "T2D3": 4,
    "T2D3H": 4,

    # 3D stress/displacement truss elements
    "T3D2": 3,
    "T3D2H": 3,
    "T3D3": 4,
    "T3D3H": 4,

    # 2D coupled temperature-displacement truss elements
    "T2D2T": 3,
    "T2D3T": 4,

    # 3D coupled temperature-displacement truss elements
    "T3D2T": 3,
    "T3D3T": 4,

    # 2D piezoelectric truss elements
    "T2D2E": 3,
    "T2D3E": 4,

    # 3D piezoelectric truss elements
    "T3D2E": 3,
    "T3D3E": 4,


#%% Beam elements
# -----------------------------------------------------------------------------

    # Beams in a plane
    "B21": 3,
    "B21H": 3,
    "B22": 4,
    "B22H": 4,
    "B23": 3,
    "B23H": 3,
    "PIPE21": 3,
    "PIPE21H": 3,
    "PIPE22": 4,
    "PIPE22H": 4,

    # Beams in space
    "B31": 3,
    "B31H": 3,
    "B32": 4,
    "B32H": 4,
    "B33": 3,
    "B33H": 3,
    "PIPE31": 3,
    "PIPE31H": 3,
    "PIPE32": 4,
    "PIPE32H": 4,

    # Open-section beams in space
    "B31OS": 3,
    "B31OSH": 3,
    "B32OS": 4,
    "B32OSH": 4,


#%% Frame elements
# -----------------------------------------------------------------------------

    # Frame in a plane
    "FRAME2D": 3,

    # Frame in space
    "FRAME3D": 4,


#%% Elbow elements
# -----------------------------------------------------------------------------

    "ELBOW31": 3,
    "ELBOW32": 4,
    "ELBOW31B": 3,
    "ELBOW31C": 3,


#%% Shear panel elements
# -----------------------------------------------------------------------------

    "SHEAR4": 12,


#%% 3d conventional shell elements
# -----------------------------------------------------------------------------

    # Stress/displacement elements
    "STRI3": 5,
    "S3": 5,
    "S3R": 5,
    "S3RS": 5,
    "STRI65": 22,
    "S4": 9,
    "S4R": 9,
    "S4RS": 9,
    "S4RSW": 9,
    "S4R5": 9,
    "S8R": 23,
    "S8R5": 23,
    "S9R5": 28,

    # Heat transfer elements
    "DS3": 5,
    "DS4": 9,
    "DS6": 22,
    "DS8": 23,

    # Coupled temperature-displacement elements
    "S3T": 5,
    "S3RT": 5,
    "S4T": 9,
    "S4RT": 9,
    "S8RT": 23,


#%% Continuum shell elements
# -----------------------------------------------------------------------------

    # Stress/displacement elements
    "SC6R": 13,
    "SC8R": 12,

    # Coupled temperature-displacement elements
    "SC6RT": 13,
    "SC8RT": 12,


#%% Axisymmetric shell elements
# -----------------------------------------------------------------------------

    # Stress/displacement elements
    "SAX1": 3,
    "SAX2": 4,

    # Heat transfer elements
    "DSAX1": 3,
    "DSAX2": 4,


# =============================================================================
#
#
#                   Inertial, rigid, capacitance elements
#
#
# =============================================================================


#%% Rigid elements
# -----------------------------------------------------------------------------

    # 2D rigid elements
    "R2D2": 3,
    "RAX2": 3,
    "RB2D2": 3,

    # 3D rigid elements
    "R3D3": 5,
    "R3D4": 9,
    "RB3D2": 3,

}



# -*- coding: utf-8 -*-

from abaqus import *
from abaqusConstants import *

import regionToolset
import math

# Open new model database
outputModel = Mdb()

modelName = 'modelTestInterface'

# Name the model
mdb.Model(name=modelName)

# Assign model and assembly to variables
model = mdb.models[modelName]
modelAssembly = model.rootAssembly

# Display assembly
session.viewports['Viewport: 1'].setValues(displayedObject=modelAssembly)


# Create an elastic material
import material
matElastic = model.Material(name='Elastic Material')
matElastic.Density(table=((0.0003,), ))
matElastic.Elastic(table=((70000.0, 0.3), ))

# Create a rectangular part
import sketch
import part

partSketch = model.ConstrainedSketch(name='Part Sketch',sheetSize=100.0)
partSketch.rectangle(point1=(0,0),point2=(1,1))
rectPart = model.Part(name='Rectangular Part', dimensionality=TWO_D_PLANAR,
                      type=DEFORMABLE_BODY)
rectPart.BaseShell(sketch=partSketch)


# Section Assignment
import section

elasticSection = model.HomogeneousSolidSection(
    name='Elastic Section',material='Elastic Material')

# Part region
regionElastic = regionToolset.Region(faces=rectPart.faces)

# Assign sections to regions
rectPart.SectionAssignment(
    region=regionElastic, sectionName='Elastic Section')

# Part set
edgeLeft = rectPart.edges.findAt(((0.0,0.5,0.0),))
rectPart.Set("Left Edge", edges=edgeLeft)

# Part surface
edgeRight = rectPart.edges.findAt(((1.0,0.5,0.0),))
rectPart.Surface("Right Edge", side1Edges=edgeRight)

# Assembly
import assembly

rectInstance = modelAssembly.Instance(name='Rectangular Instance',
    part=rectPart,dependent=ON)

# Mesh the part
import mesh

# Choose element type
elemType1 = mesh.ElemType(elemCode=CPS3, elemLibrary=STANDARD)
elemType2 = mesh.ElemType(elemCode=CPS4, elemLibrary=STANDARD)

meshFaces = rectPart.faces
meshRegion = regionToolset.Region(faces=meshFaces)

rectPart.setElementType(regions=meshRegion,elemTypes=(elemType1,elemType2))

elemSize = .5; # should create a 2x2 mesh
rectPart.seedPart(size=elemSize,constraint=FINER)
rectPart.setMeshControls(regions=rectPart.faces,technique=FREE)

rectPart.generateMesh()

# Step
import step

duration = 1
StepApplyForce = model.StaticStep(name='Apply Force',
    previous='Initial', description='Well, guess',
    nlgeom=OFF,timePeriod=duration)

Step2 = model.StaticStep(name='Another Step',
    previous='Apply Force', description='Well, guess',
    nlgeom=OFF,timePeriod=duration)

# Boundary conditions

# Inner edge fixed in r-direction
partLeftEdge = rectInstance.edges.findAt(((0.0,0.5,0.0),))
partLeftEdgeRegion = regionToolset.Region(edges=partLeftEdge)
BCLeftEdge = model.DisplacementBC(name='BC Left Edge',
    createStepName='Apply Force', region=partLeftEdgeRegion, u1=0.0)

# Bottom edge fixed in z-direction
partBottomEdge = rectInstance.edges.findAt(((0.5,0.0,0.0),))
partBottomEdgeRegion = regionToolset.Region(edges=partBottomEdge)
BCBottomEdge = model.DisplacementBC(name='BC Bottom Edge',
    createStepName='Apply Force', region=partBottomEdgeRegion, u2=0.0)
# Set for tests
modelAssembly.Set("Bottom Edge", edges=partBottomEdge)

# Force applied to top edge
partTopEdge = rectInstance.edges.findAt(((0.5,1.0,0.0),))
partTopEdgeSequence = (partTopEdge,) # Surfacetraction requires surface type
partTopEdgeRegion = modelAssembly.Surface(side1Edges=partTopEdgeSequence, name='Top Surface')

loadTraction = model.SurfaceTraction(name='Traction Top Edge',
    createStepName='Apply Force', region=partTopEdgeRegion, magnitude=10.0,
    directionVector=((0.0, 0.0, 0.0), (0.0, 1.0, 0.0)),
    distributionType=UNIFORM, field='', localCsys=None, traction=GENERAL)

# Output requests
model.FieldOutputRequest(name='FieldOut',
    createStepName='Apply Force', variables=('S', 'E', 'UT', 'COORD'),
    frequency=LAST_INCREMENT)


# Create Job
import job

jobElementTest = mdb.Job(name='element_test_current',     model=modelName,
    type=ANALYSIS, description='Uniform stress state in a rectangular body',
    numCpus=1, historyPrint=OFF)

jobElementTestOld = mdb.Job(name='element_test_old',     model=modelName,
    type=ANALYSIS, description='Uniform stress state in a rectangular body',
    numCpus=1, historyPrint=OFF)

jobElementTest.writeInput()

jobElementTestOld.writeInput()




# -*- coding: utf-8 -*-
#
#   Paraqus - A VTK exporter for FEM results.
#
#   Copyright (C) 2022, Furlan, Stollberg and Menzel
#
#    This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along with this program. If not, see https://www.gnu.org/licenses/.
"""
The main purpose of this file is to implement a model class that is
instantiated based on an abaqus odb.

One model instance is created for each time frame that is exported.

"""

# TODO: Looping over instances instead of set export requests makes it hard to
# check if all requests were executed. Right now, no error is raised when a
# set does not exist.

import os.path
import warnings

import numpy as np

from paraqus.paraqusmodel import ParaqusModel

from paraqus.abaqus.abaqustools import ODBObject
from paraqus.abaqus.elementlibrary import ABQ_ELEMENT_LIBRARY
from paraqus.constants import (ABAQUS, NODES, ELEMENTS,
                               SCALAR, TENSOR, VECTOR,
                               MEAN, ABSMAX)

from abaqusConstants import MISES, MAGNITUDE

import abaqusConstants

# lookup for invariants in field outputs
ABAQUS_INVARIANTS = {"mises": MISES,
                     "magnitude": MAGNITUDE}

# remark: surface tensors are not yet supported
PARAQUS_FIELD_TYPES = {abaqusConstants.SCALAR: SCALAR,
                       abaqusConstants.VECTOR: VECTOR,
                       abaqusConstants.TENSOR_3D_FULL: TENSOR,
                       abaqusConstants.TENSOR_3D_PLANAR: TENSOR,
                       abaqusConstants.TENSOR_2D_PLANAR: TENSOR}


class ODBReader():
    """
    Reads an odb file and exports selected results to a ParaqusModel instance.

    Attributes
    ----------
    odb_path : str
        Path to the underlying odb.
    model_name : str
        Name of the model returned by the reader.
    field_export_requests : list
        Contains one request per field that will be exported to vtk format.
    group_export_requests : list
        Contains one request per node or element set that will be exported.
    instance_names : list
        Instances that will be exported to individual models.
    time_offset : float
        Assume the simulation started at this time when exporting. Useful to
        create vtk files that are ordered in time from multiple odbs.

    Parameters
    ----------
    odb_path : str
        Name of the underlying odb.
    model_name : str, optional
        Name of the model returned by the reader. When `model_name` is omitted,
        it is set based on `odb_path`. Default: None.
    instance_names : list, optional
        Instances that will be exported to individual models. If
        `instance_names` is omitted, all instances in the odb will be exported.
        Default: None.
    time_offset : float, optional
        Assume the simulation started at this time when exporting. Useful to
        create vtk files that are ordered in time from multiple odbs.
        Default: 0.0

    Examples
    --------
    >>> from paraqus.abaqus import ODBReader
    >>> # Define the odb/model/instances that will be read
    >>> reader = ODBReader(odb_path="example.odb",
    >>>                    model_name="example model",
    >>>                    instance_names=["instance-1", "instance-2"],
    >>>                    )
    >>> # Export the node field 'U' (displacements)
    >>> reader.add_field_export_request("U", field_position="nodes")
    >>> # export the element set 'element set name' only for 'instance-1'
    >>> reader.add_set_export_request("element set name",
    >>>                               set_type="elements",
    >>>                               instance_name="instance-1")
    >>> # store the models for both instances in a list
    >>> instance_models = list(reader.read_instances(step_name="Step-1",
    >>>                                              frame_index=-1)
    >>>                       )

    """

    def __init__(self,
                 odb_path,
                 model_name=None,
                 instance_names=None,
                 time_offset=0.0):

        self.odb_path = odb_path
        self.model_name = model_name
        self.field_export_requests = []
        self.group_export_requests = []
        self.instance_names = instance_names
        self.time_offset = time_offset


    def get_number_of_frames(self, step_name):
        """
        Return the number of frames for a given step in the underlying odb.

        The reader does NOT check that all output is available in each frame.

        Parameters
        ----------
        step_name : str
            Name of the step in the odb.

        Returns
        -------
        number_of_frames : int
            Number of frames in the step, including the initial (0th) frame.

        """
        with ODBObject(self.odb_path) as odb:
            step = odb.steps[step_name]
            return len(step.frames)


    def get_frame_indices(self, step_name, how='all'):
        """
        Return the indices of the frames with at least one requested fiels.
        
        Parameters
        ----------
        step_name : str
            Name of the step in the odb.
        how : str
            Whether to get frame indices where 'any' or 'all' fields have
            values. Default: 'all'.
        
        Returns
        -------
        indices : List[int]
        
        """
        if len(self.field_export_requests) == 0:
            raise RuntimeError(
                "Can not check for frame indices when no field export "
                "requests are registered.")
        
        # names of the fields
        field_names = [er.field_name for er in self.field_export_requests]
        
        # function to check if at least one field is present in a frame
        if how=='all':
            reduction = all
        elif how=='any':
            reduction = any
        else:
            raise ValueError("'how' argument must be 'any' or 'all'.")
        
        def has_data(f):
            return reduction([n in f.fieldOutputs for n in field_names])
                
                
        with ODBObject(self.odb_path) as odb:
            step = odb.steps[step_name]
            frames = step.frames
                        
            frame_indices = [i for (i,f) in enumerate(frames) if has_data(f)]
        
        return frame_indices


    def add_field_export_request(self,
                                 field_name,
                                 **kwargs):
        """
        Request export of an output field.

        Parameters
        ----------
        field_name : str
            Abaqus identifier for the field, e.g. 'S' for stress.
        **kwargs : dict
            Additional parameters that will be passed to the underlying
            FieldExportRequest.

        Returns
        -------
        None

        """
        request = FieldExportRequest(field_name, **kwargs)

        self.field_export_requests.append(request)

    def add_set_export_request(self,
                               set_name,
                               set_type,
                               instance_name=None):
        """
        Request that a node or element set is exported.

        Parameters
        ----------
        set_name : str
            Name of the set in the odb
        set_type : str
            Type of the set. Valid values are 'nodes' or 'elements'.
        instance_name : str, optional
            If `instance_name` is not None, look for an instance-level set
            in the odb. Otherwise look for an assembly-level set.
            Default: None.

        Returns
        -------
        None

        """
        request = GroupExportRequest(set_name, set_type, instance_name)

        self.group_export_requests.append(request)


    def add_surface_export_request(self,
                                   surface_name,
                                   surface_type,
                                   instance_name=None):
        """
        Request that a node or element set is exported.

        Parameters
        ----------
        surface_name : str
            Name of the surface in the odb
        surface_type : str
            Type of the surface. Valid values are 'nodes' or 'elements'.
        instance_name : str, optional
            If `instance_name` is not None, look for an instance-level surface
            in the odb. Otherwise look for an assembly-level surface.
            Default: None.

        Returns
        -------
        None

        """
        request = GroupExportRequest(surface_name,
                                     surface_type,
                                     instance_name,
                                     surface_set=True)

        self.group_export_requests.append(request)


    def get_frame_time(self, step_name, frame_index):
        """
        Get the highest time value of any frame in the odb.

        The `time_offset` specified for the reader is added to the value.

        Parameters
        ----------
        step_name : str
            Name of the step in Abaqus.
        frame_index : int
            Index of the frame in the step in Abaqus.

        Returns
        -------
        frame_time : float
            The total time passed (includes time offset).

        """
        with ODBObject(self.odb_path) as odb:
            step = odb.steps[step_name]
            frame = step.frames[frame_index]
            frame_time =  self.time_offset + step.totalTime + frame.frameValue

        return frame_time


    def read_instances(self, step_name, frame_index):
        """
        Read a frame from the underlying odb.

        Each instance is read separately and returns an individual
        ParaqusModel. Models are generated lazily, so memory-efficient
        iteration over the return values is possible.

        Parameters
        ----------
        step_name : str
            Name of the step in Abaqus.
        frame_index : int
            Index of the frame in the step in Abaqus.

        Yields
        ------
        model : ParaqusModel
            Model representing the geometry and output for one part instance
            at the step/frame.

        """
        if self.model_name is None:
            model_name =  os.path.splitext(os.path.basename(self.odb_path))[0]
        else:
            model_name = self.model_name

        with ODBObject(self.odb_path) as odb:
            # TODO: Input checking
            step = odb.steps[step_name]
            frame = step.frames[frame_index]

            frame_time =  self.time_offset + step.totalTime + frame.frameValue

            # if no instances are specified, export all of them
            if self.instance_names is None:
                instance_names = list(odb.rootAssembly.instances.keys())
            else:
                instance_names = self.instance_names


            # loop over the requested instances
            for instance_name in instance_names:

                instance = odb.rootAssembly.instances[instance_name]

                # extract nodes/elements and instance-level sets
                mesh = InstanceMesh(instance)

                # create a model object based on the mesh info
                model = ParaqusModel(mesh.element_labels,
                                     mesh.element_connectivities,
                                     mesh.element_types,
                                     mesh.node_labels,
                                     mesh.node_coords,
                                     model_name=model_name,
                                     part_name=instance_name,
                                     step_name=step_name,
                                     frame_time=frame_time,
                                     source=ABAQUS)

                # loop to add field outputs
                for request in self.field_export_requests:
                    fo = self._get_field_output(request,
                                                frame,
                                                instance=instance)

                    if fo is None:
                        # skip empty outputs when exporting
                        msg = ("Field output {} ".format(request.field_name)
                              + "not available in instance "
                              + "{}. ".format(instance_name)
                              + "Export skipped." )
                        warnings.warn(msg)
                        continue

                    (field_tags,
                     field_data,
                     field_position,
                     field_type,
                     export_name) = self._read_field_output(request, fo, mesh)

                    model.add_field(export_name,
                                    field_tags,
                                    field_data,
                                    field_position,
                                    field_type)

                # loop to export node/element sets
                for request in self.group_export_requests:
                    group_tags = self._get_group_tags(request, odb, instance)

                    # do nothing if the request is not for this instance
                    if group_tags is None:
                        continue

                    # add the actual groups to the model
                    if request.group_type == NODES:
                        model.nodes.add_group(request.export_name,
                                              group_tags)
                    elif request.group_type == ELEMENTS:
                        model.elements.add_group(request.export_name,
                                                 group_tags)
                    else:
                        raise ValueError(
                            "Request type must be 'nodes' or 'elements'.")

                yield model


    def _get_group_tags(self, request, odb, instance):
        """
        Return the node/element tags for a node/element set.

        Parameters
        ----------
        request : GroupExportRequest
            Describes a set or surface that will be exported.
        odb : Abaqus ODB object
            The open odb.
        instance : Abaqus part instance object
            The part instance that is currently exported.

        Returns
        -------
        tags : ArrayLike or None
            Node/element numbers in the surface or set. If the export request
            is for another instance, None is returned.

        """
        # we will read the attributes based on the group type
        if request.group_type == NODES:
            group_type = "nodes"
            set_type = "nodeSets"
        elif request.group_type == ELEMENTS:
            group_type = "elements"
            set_type = "elementSets"


        if request.instance_name is None:
            # assembly-level set/surface

            # extract an odbSet representing the surface or set
            if request.surface_set:
                # the surface returns as an odbSet directly
                set_object = odb.rootAssembly.surfaces[request.group_name]
            else:
                # for node/element sets, we need to take one more step
                set_repository = odb.rootAssembly.__getattribute__(set_type)
                set_object = set_repository[request.group_name]

            # for assembly-level sets or surfaces, the nodes/elements are
            # stored per instance, so we extract the index for the current
            # instance
            try:
                index = set_object.instances.index(instance)
            except ValueError:
                # instance has no nodes/elements in the set
                return np.array([], dtype=int)

            # extract the nodes/elements as an odbMeshNodeArray or
            # odbMeshElementArray. This array has only nodes/elements of
            # the instance under consideration
            array_object = set_object.__getattribute__(group_type)[index]

            labels = np.array([o.label for o in array_object])

            # sort the array
            labels.sort()

            # remove duplicates (these do actually occur)
            return np.unique(labels)

        elif request.instance_name == instance.name:
            # instance-level set/surface

            # extract an odbSet representing the surface or set
            if request.surface_set:
                # the surface returns as an odbSet directly
                set_object = instance.surfaces[request.group_name]
            else:
                # for node/element sets, we need to take one more step
                set_repository = instance.__getattribute__(set_type)
                set_object = set_repository[request.group_name]

            # extract the nodes/elements as an odbMeshNodeArray or
            # odbMeshElementArray
            array_object = set_object.__getattribute__(group_type)

            # return an array of node or element labels
            labels = np.array([o.label for o in array_object])

            # sort the array
            labels.sort()

            # remove duplicates (these do actually occur)
            return np.unique(labels)

        else:
            # The request was not for this instance
            return None


    def _get_field_output(self,
                          request,
                          frame,
                          instance=None):
        """Get a field output object for the requested field."""
        field_name = request.field_name
        invariant = request.invariant

        # get the field output for the current field
        field_out = frame.fieldOutputs[field_name]

        # reduce to the instance
        if instance is not None:
            field_out = field_out.getSubset(region=instance)

        # output should be uniform in terms of location (node or qp output)
        # FIXME: This is propably not needed anymore
        # assert len(field_out.locations) == 1

        # if an invariant is specified, apply the reduction
        if invariant is not None:
            abq_invar = ABAQUS_INVARIANTS[invariant.lower().strip()]
            field_out = field_out.getScalarField(invariant=abq_invar)

        # if a field position is specified, reduce the output correspondingly
        if request.field_position is not None:
            if request.field_position == NODES:
                field_out = field_out.getSubset(position=abaqusConstants.NODAL)
            elif request.field_position == ELEMENTS:
                field_out = field_out.getSubset(position=abaqusConstants.CENTROID)
            else:
                raise ValueError("Position not implemented.")

        if len(field_out.values) == 0:
            # no data for the field at this position
            return None

        return field_out


    def _read_field_output(self, request, field_out, instance_mesh):
        """
        Read the data for one field output.

        Parameters
        ----------
        request : FieldExportRequest
            Specifies which field will be read from the odb.
        field_out : Abaqus FieldOutput
            Field output corresponding to the request (may still have values
            at other points than requested etc).
        instance_mesh : InstanceMesh
            Mesh representation of the instance the output is requested for,
            used to extract the correct values.

        Returns
        -------
        labels : ArrayLike
            Node or element labels.
        data : ArrayLike
            Field values in the same order as `labels`.
        paraqus_position
            Specifies where the output is evaluated ('nodes' or 'elements').
        field_type
            What type of field (scalar/vector/tensor).

        """
        blocks = field_out.bulkDataBlocks

        position = blocks[0].position

        assert all([b.position == position for b in blocks]), \
            "All data for one field must exist at the same output position."

        # make sure element output is actually per element
        if position in (abaqusConstants.INTEGRATION_POINT,
                        abaqusConstants.ELEMENT_NODAL):
            field_out = field_out.getSubset(position=abaqusConstants.CENTROID)

            blocks = field_out.bulkDataBlocks
            position = blocks[0].position

        # possible values for the position at this point:
        # NODAL, CENTROID, WHOLE_ELEMENT
        if position == abaqusConstants.NODAL:
            paraqus_position = NODES

            (labels,
             data,
             field_type,
             field_description) = self._get_node_data(request,
                                                      field_out,
                                                      instance_mesh)

        elif position in (abaqusConstants.CENTROID,
                          abaqusConstants.WHOLE_ELEMENT):
            paraqus_position = ELEMENTS

            (labels,
             data,
             field_type,
             field_description) = self._get_element_data(request,
                                                         field_out,
                                                         instance_mesh)
        else:
            raise ValueError("Position not implemented.")

        # change 5th and 6th component to be compatible to vtk element order
        if field_type == TENSOR and data.shape[1] == 6:
            data = data[:,[0,1,2,3,5,4]]

        return labels, data, paraqus_position, field_type, field_description


    def _get_node_data(self, request, field_out, instance_mesh):
        """Extract node values from a FieldOutput."""
        instance_nodes = instance_mesh.node_labels

        blocks = field_out.bulkDataBlocks
        dtype = blocks[0].data.dtype
        ncomponents = blocks[0].data.shape[1]
        field_type = PARAQUS_FIELD_TYPES[blocks[0].type]

         # initialize data as NaNs
        data = np.ones((len(instance_nodes), ncomponents), dtype=dtype)*np.nan

        # create a description for the field
        description_str = request.field_name

        if request.invariant is not None:
            description_str += "_{}".format(request.invariant)

        # for each node there is only a single value
        for block in blocks:
            block_nodes = block.nodeLabels
            block_data = block.data

            # indices of the block nodes in instance_nodes
            # this assumes that instance_nodes is sorted!
            indices = np.searchsorted(instance_nodes, block_nodes)

            data[indices,:] = block_data

        return instance_nodes, data, field_type, description_str


    def _get_element_data(self, request, field_out, instance_mesh):
        """Extract element values from a FieldOutput."""
        instance_elements = instance_mesh.element_labels

        blocks = field_out.bulkDataBlocks
        dtype = blocks[0].data.dtype
        ncomponents = blocks[0].data.shape[1]
        field_type = PARAQUS_FIELD_TYPES[blocks[0].type]

        # section point number for every block
        # since abaqus numbers section points from 1, we use 0 for None values
        section_points = np.zeros(len(blocks), dtype=int)

        for ib, block in enumerate(blocks):
            if block.sectionPoint is not None:
                section_points[ib] = block.sectionPoint.number

        # mapping section point number -> index in data array
        section_point_indices = {n: i
             for (i,n) in enumerate(np.unique(section_points))}

        # assume these is only one data value for each (element, sectionPoint)
        nel = len(instance_elements)
        nsp = len(section_point_indices)

        # initialize data as NaNs
        data = np.ones((nel, nsp, ncomponents), dtype=dtype)*np.nan

        # create a description for the field
        description_str = request.field_name

        if request.invariant is not None:
            description_str += "_{}".format(request.invariant)

        if request.section_point_number is not None:
            description_str += "_sp{}".format(request.section_point_number)
        elif nsp > 1:
            # triggers only if there is no explicit specification of the
            # section point
            description_str += "_{}".format(request.section_point_reduction)

        # extract the data by looping over the data blocks
        for block, sp_number in zip(blocks, section_points):
            sp_index = section_point_indices[sp_number]

            block_elements = block.elementLabels
            block_data = block.data

            assert len(np.setdiff1d(block_elements, instance_elements)) == 0, \
                "not all block element labels are part of the instance."

            # indices iof the block elements in instance_elements
            # this assumes that instance_elements is sorted!
            el_indices = np.searchsorted(instance_elements, block_elements)

            data[el_indices,sp_index,:] = block_data

        # if a specific section point was requested:
        if request.section_point_number is not None:
            assert request.section_point_number in section_point_indices, \
                "Requested section point number not in output."

            sp_index = section_point_indices[request.section_point_number]
            return (instance_elements,
                    data[:,sp_index,:],
                    field_type,
                    description_str)

        if nsp == 1:
            # "standard" case where there are not section points (or only 1)
            return (instance_elements,
                    data[:,0,:],
                    field_type,
                    description_str)
        else:
            # reductions for multiple section points
            if request.section_point_reduction == MEAN:
                # average data ignoring nans
                data = np.nanmean(data, axis=1)
            elif request.section_point_reduction == ABSMAX:
                data = np.nanmax(np.abs(data), axis=1)
            else:
                raise ValueError("Reduction method invalid.")

            return (instance_elements,
                    data,
                    field_type,
                    description_str)


class InstanceMesh():
    """
    Represent the mesh for one part instance.

    Attributes
    ----------
    node_labels : ArrayLike
        Node numbers for the part instance.
    node_coords : ArrayLike
        Coordinates of the nodes.
    element_labels : ArrayLike
        Element number for each element in the instance.
    element_types : ArrayLike
        VTK element type for each element, in the same order as
        `element_labels`.
    element_connectivities : List[Tuple[int]]
        Nodes in each element, in the same order as `element_labels`.

    Parameters
    ----------
    instance : Abaqus odb instance object
        The instance the mesh is extracted from.
    sort_values : Bool, optional
        Whether the nodes and elements are sorted by label. Default: True.

    """

    def __init__(self, instance, sort_values=True):

        self._read_nodes(instance, sort=sort_values)
        self._read_elements(instance, sort=sort_values)
        self._read_node_sets(instance)
        self._read_element_sets(instance)

    def _read_nodes(self, instance, sort=True):
        """
        Extract node info.

        Sets the attributes `node_tags` and `node_coords`.

        Parameters
        ----------
        instance : abaqus odb instance object
            The instance in question.
        sort : bool, optional
            Whether to sort the return by node label. Default: True.

        """
        labels, coords = zip(*[(n.label, n.coordinates)
                               for n in instance.nodes])

        labels = np.array(labels, dtype=int)
        coords = np.array(coords)

        # sort output by node label
        if sort:
            sorter = np.argsort(labels)
            labels = labels[sorter]
            coords = coords[sorter]

        self.node_labels = labels
        self.node_coords = coords


    def _read_elements(self, instance, sort=True):
        """
        Extract element info.

        Sets the attributes `element_tags` , `element_types` and
        `element_connectivities`.

        Parameters
        ----------
        instance : abaqus odb instance object
            The instance in question.
        sort : bool, optional
            Whether to sort the return by element label. Default: True.

        """
        # using zip as its own inverse is faster than multiple comprehensions
        labels, types, connectivities = zip(*[(e.label, e.type, e.connectivity)
                                              for e in instance.elements])

        labels = np.array(labels)

        # conversion to integer vtk element codes
        types = np.array([ABQ_ELEMENT_LIBRARY[str(el_type)]
                          for el_type in types],
                         dtype=int)

        # since connectivities are tuples of possibly different lengths,
        # we do not convert these to arrays

        # sort output by element label
        if sort:
            sorter = np.argsort(labels)

            labels = labels[sorter]
            types = types[sorter]

            # convert back to tuples so return type stays the same
            # types = tuple([types[i] for i in sorter])

            connectivities = [tuple(connectivities[i]) for i in sorter]

        self.element_labels = labels
        self.element_types = types
        self.element_connectivities = connectivities


    def _read_node_sets(self, instance):
        """
        Extract instance-level node sets as indicator arrays.

        Sets the attribute `node_sets`.

        Parameters
        ----------
        instance : abaqus odb instance object
            The instance in question.

        """
        self.node_sets = {}

        for set_name, nset in instance.nodeSets.items():
            set_labels = np.array([n.label for n in nset.nodes])
            set_indicators = np.array([label in set_labels
                                       for label in self.node_labels])

            self.node_sets[set_name] = set_indicators.astype(int)


    def _read_element_sets(self, instance):
        """
        Extract instance-level element sets as indicator arrays.

        Sets the attribute `element_sets`.

        Parameters
        ----------
        instance : abaqus odb instance object
            The instance in question.

        """
        self.element_sets = {}

        for set_name, eset in instance.elementSets.items():
            set_labels = np.array([e.label for e in eset.elements])
            set_indicators = np.array([label in set_labels
                                       for label in self.element_labels])

            self.element_sets[set_name] = set_indicators.astype(int)


class FieldExportRequest():
    """
    Specify a field output that will be exported.

    Attributes
    ----------
    See parameters.

    Parameters
    ----------
    field_name : str
        Abaqus identifier for the field, e.g. 'S' for stress.
    field_position : str, optional
        What type of field. valid values: 'nodes' or 'elements'.
    invariant : str, optional
        Invariant of the field, if applicable. E.g. 'mises', 'magnitude'.
    section_point_number : int, optional
        Section point for shell elements. If no section point number is
        specified, the section point values are reduced according to
        `section_point_reduction`.
    section_point_reduction : str
        How to reduce values of multiple section points for the same node or
        element. Valid values are 'mean' or 'absmax'.

    """

    def __init__(self,
                 field_name,
                 field_position=None,
                 invariant=None,
                 section_point_number=None,
                 section_point_reduction="absmax"):

        self.field_name = field_name
        self.field_position = field_position
        self.invariant = invariant
        self.section_point_number = section_point_number

        assert section_point_reduction in (MEAN, ABSMAX), \
            "Valid values for section_point_reduction are 'mean' or 'absmax'."

        self.section_point_reduction = section_point_reduction


class GroupExportRequest():
    """
    Specify a group of nodes or elements that will be written to the vtk file.

    Attributes
    ----------
    See parameters

    Parameters
    ----------
    group_name : str
        Name of the set or surface in the odb
    group_type : str
        Type of the set. valid values are 'nodes' or 'elements'.
    instance_name : str, optional
        If `instance_name` is not None, look for an instance-level set
        in the odb. Otherwise look for an assembly-level set.
        Default: None.
    surface_set : bool, optional
        If `False`, export a set. If `True`, export nodes/elements of a
        surface.

    """

    def __init__(self,
                 group_name,
                 group_type,
                 instance_name=None,
                 surface_set=False):

        self.group_name = group_name
        self.group_type = group_type
        self.instance_name = instance_name
        self.surface_set = surface_set

        assert group_type in (NODES, ELEMENTS), \
            "Valid values for group_type are 'nodes' or 'elements'."

    @property
    def export_name(self):
        """Return a unique name for the surface/set that is exported."""
        export_name = ""
        if self.surface_set:
            export_name += "surface."

        if self.instance_name is not None:
            export_name += self.instance_name + "."

        export_name += self.group_name

        return export_name



"""
Tests for the ParaqusModel class.

These tests can be executed in Abaqus python or in standard python 2.7+.

"""
import unittest
import numpy as np

from paraqus.paraqusmodel import ParaqusModel

class TestParaqusModelCreation(unittest.TestCase):

    def test_model_creation(self):
        """A ParaqusModel can be created based on mesh info."""
        # Define geometry consisting of two first-order quad elements and
        # three first-order tri elements
        node_tags = [1, 2, 3, 4, 5, 6, 7, 8]
        node_coords = [[0, 0],
                       [1, 0],
                       [2, 0],
                       [0, 1],
                       [1, 1],
                       [2, 1],
                       [0.5, 1.5],
                       [1.5, 1.5]]
        element_tags = [1, 2, 3, 4, 5]
        connectivity = [[1, 2, 5, 4],
                        [2, 3, 6, 5],
                        [4, 5, 7],
                        [5, 6, 8],
                        [5, 8, 7]]
        element_types = [9, 9, 5, 5, 5]

        # Create four element base model
        model_name = "2D_TEST_MODEL"
        part_name = "2D_TEST_PART"
        model = ParaqusModel(element_tags,
                             connectivity,
                             element_types,
                             node_tags,
                             node_coords,
                             model_name=model_name,
                             part_name=part_name)

        assert model.model_name == model_name
        assert model.part_name == part_name
        assert len(model.nodes.tags) == 8
        assert len(model.elements.tags) == 5


class TestParaqusModelFields(unittest.TestCase):

    def setUp(self):
        """Create a small ParaqusModel."""
        # Define geometry consisting of two first-order quad elements and
        # three first-order tri elements
        node_tags = [1, 2, 3, 4, 5, 6, 7, 8]
        node_coords = [[0, 0],
                       [1, 0],
                       [2, 0],
                       [0, 1],
                       [1, 1],
                       [2, 1],
                       [0.5, 1.5],
                       [1.5, 1.5]]
        element_tags = [1, 2, 3, 4, 5]
        connectivity = [[1, 2, 5, 4],
                        [2, 3, 6, 5],
                        [4, 5, 7],
                        [5, 6, 8],
                        [5, 8, 7]]
        element_types = [9, 9, 5, 5, 5]

        # Create four element base model
        model_name = "2D_TEST_MODEL"
        part_name = "2D_TEST_PART"
        self.model = ParaqusModel(element_tags,
                                  connectivity,
                                  element_types,
                                  node_tags,
                                  node_coords,
                                  model_name=model_name,
                                  part_name=part_name)

    def test_scalar_node_field(self):
        """Scalar node fields can be added to models, value order is ok."""
        field_tags = [1, 3, 2, 4, 5, 6, 7, 8]
        field_vals = [1, 2, 3, 4, 5, 6, 7, 8]

        self.model.add_field("scalar node field",
                             field_tags,
                             field_vals,
                             "nodes",
                             "scalar")

        field = self.model.get_node_field("scalar node field")

        assert field.field_name == "scalar node field"
        assert field.field_type == "SCALAR"
        assert field.field_position == "NODES"
        assert np.all(field.field_values == [[1],[3],[2],[4],[5],[6],[7],[8]])

    def test_vector_node_field(self):
        """Vector node fields can be added to models, value order is ok."""
        field_tags = [1,3,2,4,5,6,7,8]
        field_vals = [[11, 12],
                      [21, 22],
                      [31, 32],
                      [41, 42],
                      [51, 52],
                      [61, 62],
                      [71, 72],
                      [81, 82]]

        self.model.add_field("vector node field",
                             field_tags,
                             field_vals,
                             "nodes",
                             "vector")

        field = self.model.get_node_field("vector node field")

        assert field.field_name == "vector node field"
        assert field.field_type == "VECTOR"
        assert field.field_position == "NODES"
        assert np.all(field.field_values == [[11, 12],
                                             [31, 32],
                                             [21, 22],
                                             [41, 42],
                                             [51, 52],
                                             [61, 62],
                                             [71, 72],
                                             [81, 82]])

    def test_tensor_node_field(self):
        """Tensor node fields can be added to models, value order is ok."""
        field_tags = [1, 3, 2, 4, 5, 6, 7, 8]
        field_vals = [[11, 12, 13, 14],
                      [21, 22, 23, 24],
                      [31, 32, 33, 34],
                      [41, 42, 43, 44],
                      [51, 52, 53, 54],
                      [61, 62, 63, 64],
                      [71, 72, 73, 74],
                      [81, 82, 83, 84]]

        self.model.add_field("tensor node field",
                             field_tags,
                             field_vals,
                             "nodes",
                             "tensor")

        field = self.model.get_node_field("tensor node field")

        assert field.field_name == "tensor node field"
        assert field.field_type == "TENSOR"
        assert field.field_position == "NODES"
        assert np.all(field.field_values == [[11, 12, 13, 14],
                                             [31, 32, 33, 34],
                                             [21, 22, 23, 24],
                                             [41, 42, 43, 44],
                                             [51, 52, 53, 54],
                                             [61, 62, 63, 64],
                                             [71, 72, 73, 74],
                                             [81, 82, 83, 84]])

    def test_scalar_element_field(self):
        """Scalar element fields can be added to models, value order is ok."""
        field_tags = [1, 3, 2, 4, 5]
        field_vals = [1, 2, 3, 4, 5]

        self.model.add_field("scalar element field",
                             field_tags,
                             field_vals,
                             "elements",
                             "scalar")

        field = self.model.get_element_field("scalar element field")

        assert field.field_name == "scalar element field"
        assert field.field_type == "SCALAR"
        assert field.field_position == "ELEMENTS"
        assert np.all(field.field_values == [[1],[3],[2],[4],[5]])

    def test_vector_element_field(self):
        """Vector element fields can be added to models, value order is ok."""
        field_tags = [1,3,2,4,5]
        field_vals = [[11, 12],
                      [21, 22],
                      [31, 32],
                      [41, 42],
                      [51, 52]]

        self.model.add_field("vector element field",
                             field_tags,
                             field_vals,
                             "elements",
                             "vector")

        field = self.model.get_element_field("vector element field")

        assert field.field_name == "vector element field"
        assert field.field_type == "VECTOR"
        assert field.field_position == "ELEMENTS"
        assert np.all(field.field_values == [[11, 12],
                                             [31, 32],
                                             [21, 22],
                                             [41, 42],
                                             [51, 52]])

    def test_tensor_element_field(self):
        """Tensor element fields can be added to models, value order is ok."""
        field_tags = [1, 3, 2, 4, 5]
        field_vals = [[11, 12, 13, 14],
                      [21, 22, 23, 24],
                      [31, 32, 33, 34],
                      [41, 42, 43, 44],
                      [51, 52, 53, 54]]

        self.model.add_field("tensor element field",
                             field_tags,
                             field_vals,
                             "elements",
                             "tensor")

        field = self.model.get_element_field("tensor element field")

        assert field.field_name == "tensor element field"
        assert field.field_type == "TENSOR"
        assert field.field_position == "ELEMENTS"
        assert np.all(field.field_values == [[11, 12, 13, 14],
                                             [31, 32, 33, 34],
                                             [21, 22, 23, 24],
                                             [41, 42, 43, 44],
                                             [51, 52, 53, 54]])

    def test_pad_scalar_field_values(self):
        """Missing values in scalar fields are padded with Nans."""
        # no value for node 4
        field_tags = [1, 3, 2, 5]
        field_vals = [1, 2, 3, 5]

        self.model.add_field("scalar element field",
                             field_tags,
                             field_vals,
                             "elements",
                             "scalar")

        field = self.model.get_element_field("scalar element field")

        assert field.field_name == "scalar element field"
        assert field.field_type == "SCALAR"
        assert field.field_position == "ELEMENTS"
        np.testing.assert_array_equal(field.field_values, [[1],
                                                           [3],
                                                           [2],
                                                           [np.nan],
                                                           [5]])

    def test_pad_vector_field_values(self):
        """Missing values in vector fields are padded with Nans."""
        # no value for node 4
        field_tags = [1, 3, 2, 5]
        field_vals = [[11, 12, 13],
                      [31, 32, 33],
                      [21, 22, 23],
                      [51, 52, 53]]

        self.model.add_field("vector element field",
                             field_tags,
                             field_vals,
                             "elements",
                             "vector")

        field = self.model.get_element_field("vector element field")

        assert field.field_name == "vector element field"
        assert field.field_type == "VECTOR"
        assert field.field_position == "ELEMENTS"
        np.testing.assert_array_equal(field.field_values, [[11, 12, 13],
                                                           [21, 22, 23],
                                                           [31, 32, 33],
                                                           [np.nan]*3,
                                                           [51, 52, 53]])

    def test_error_scalar_field_multiple_components(self):
        """An exception is raised when data for a scalar field is not 1d."""
        field_tags = [1, 2, 3, 4, 5]
        field_vals = [[1, 1], [2, 2], [3, 3], [4, 4], [5, 5]]

        with self.assertRaises(ValueError) as context:
            self.model.add_field("scalar element field",
                             field_tags,
                             field_vals,
                             "elements",
                             "scalar")





class TestParaqusModelSplitting(unittest.TestCase):

    def setUp(self):
        """Create a small ParaqusModel."""
        # Define geometry consisting of two first-order quad elements and
        # three first-order tri elements
        node_tags = [1, 2, 3, 4, 5, 6, 7, 8]
        node_coords = [[0, 0],
                       [1, 0],
                       [2, 0],
                       [0, 1],
                       [1, 1],
                       [2, 1],
                       [0.5, 1.5],
                       [1.5, 1.5]]
        element_tags = [1, 2, 3, 4, 5]
        connectivity = [[1, 2, 5, 4],
                        [2, 3, 6, 5],
                        [4, 5, 7],
                        [5, 6, 8],
                        [5, 8, 7]]
        element_types = [9, 9, 5, 5, 5]

        # Create four element base model
        model_name = "2D_TEST_MODEL"
        part_name = "2D_TEST_PART"
        self.model = ParaqusModel(element_tags,
                                  connectivity,
                                  element_types,
                                  node_tags,
                                  node_coords,
                                  model_name=model_name,
                                  part_name=part_name)



    def test_partition_model(self):
        """The model can be split into multiple submodels."""
        num_models = 2

        submodels = list(self.model.split_model(num_models))

        assert len(submodels) == 2

        assert all([len(m.elements.tags) > 0 for m in submodels])

        etags1 = submodels[0].elements.tags
        etags2 = submodels[1].elements.tags

        # elements can only be in one submodel!
        assert set(self.model.elements.tags) == set(etags1) ^ set(etags2)

        ntags1 = submodels[0].nodes.tags
        ntags2 = submodels[1].nodes.tags

        # nodes can be shared
        assert set(self.model.nodes.tags) == set(ntags1) | set(ntags2)



class TestParaqusModelFieldAccess(unittest.TestCase):

    def test_fields_by_type(self):
        """Fields stored in the model can be accessed via their type."""

        # Define geometry consisting of two first-order quad elements and
        # three first-order tri elements
        node_tags = [1, 2, 3, 4, 5, 6, 7, 8]
        node_coords = [[0, 0],
                       [1, 0],
                       [2, 0],
                       [0, 1],
                       [1, 1],
                       [2, 1],
                       [0.5, 1.5],
                       [1.5, 1.5]]
        element_tags = [1, 2, 3, 4, 5]
        connectivity = [[1, 2, 5, 4],
                        [2, 3, 6, 5],
                        [4, 5, 7],
                        [5, 6, 8],
                        [5, 8, 7]]
        element_types = [9, 9, 5, 5, 5]

        # Create four element base model
        model_name = "2D_TEST_MODEL"
        part_name = "2D_TEST_PART"
        self.model = ParaqusModel(element_tags,
                                  connectivity,
                                  element_types,
                                  node_tags,
                                  node_coords,
                                  model_name=model_name,
                                  part_name=part_name)

        scalar_node_vals = [1, 2, 3, 4, 5, 6, 7, 8]

        self.model.add_field("scalar node field",
                             node_tags,
                             scalar_node_vals,
                             "nodes",
                             "scalar")

        vector_node_vals = [[11, 12],
                            [21, 22],
                            [31, 32],
                            [41, 42],
                            [51, 52],
                            [61, 62],
                            [71, 72],
                            [81, 82]]

        self.model.add_field("vector node field",
                             node_tags,
                             vector_node_vals,
                             "nodes",
                             "vector")

        scalar_element_vals = [1, 2, 3, 4, 5]

        self.model.add_field("scalar element field",
                             element_tags,
                             scalar_element_vals,
                             "elements",
                             "scalar")

        vector_element_vals = [[11, 12],
                               [21, 22],
                               [31, 32],
                               [41, 42],
                               [51, 52]]

        self.model.add_field("vector element field",
                             element_tags,
                             vector_element_vals,
                             "elements",
                             "vector")

        node_fields_scalar = self.model.get_fields_by_type("scalar", "nodes")
        assert len(node_fields_scalar) == 1
        assert node_fields_scalar[0]._field_type == "scalar"
        assert node_fields_scalar[0]._field_position == "nodes"

        node_fields_vector = self.model.get_fields_by_type("vector", "nodes")
        assert len(node_fields_vector) == 1
        assert node_fields_vector[0]._field_type == "vector"
        assert node_fields_vector[0]._field_position == "nodes"

        element_fields_scalar = self.model.get_fields_by_type("scalar", "elements")
        assert len(element_fields_scalar) == 1
        assert element_fields_scalar[0]._field_type == "scalar"
        assert element_fields_scalar[0]._field_position == "elements"

        element_fields_vector = self.model.get_fields_by_type("vector", "elements")
        assert len(element_fields_vector) == 1
        assert element_fields_vector[0]._field_type == "vector"
        assert element_fields_vector[0]._field_position == "elements"



# -*- coding: utf-8 -*-
#
#   Paraqus - A VTK exporter for FEM results.
#
#   Copyright (C) 2022, Furlan, Stollberg and Menzel
#
#    This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along with this program. If not, see https://www.gnu.org/licenses/.
"""
This module containts all classes that contribute to paraqus models,
i.e. not just the model class itself but also repositories for nodes,
elements and fields.

"""
import itertools
import warnings
import numpy as np
from abc import ABCMeta, abstractmethod

import paraqus.constants as constants
from paraqus.constants import USER, NODES, ELEMENTS, SCALAR, VECTOR, TENSOR


class ParaqusModel(object):
    """
    Paraqus representation of a finite element model.

    The paraqus model can be used to manipulate and export any finite
    element model as a set of VTK files.

    Parameters
    ----------
    element_tags : Sequence
        Tags of all elements being part of the model. Tags have to be
        unique.
    connectivity : Sequence of Sequence
        Connectivity list in order of the element tags. The respective
        nodes have to be in order as defined for VTK cells.
    element_types : Sequence
        The VTK cell types for all elements in order of the element
        tags.
    node_tags : Sequence
        Tags of all nodes being part of the model. Tags have to be
        unique.
    node_coords : Sequence of Sequence
        Nodal coordinates in order of the node tags. The coordinates
        can be defined in 1d-, 2d- or 3d-space as an array of shape
        (number of nodes x number of dimensions).
    model_name : str, optional
        Name of the main model. In case of multiple frames or parts
        every paraqus model should have the same model name. Default is
        'MODEL NAME'.
    part_name : str, optional
        Name of the part instance. Default is 'PART NAME'.
    step_name : str, optional
        Name of the load step. Default is 'STEP NAME'.
    frame_time : float, optional
        Current frame time. Default is 0.0.
    source : ParaqusConstant, optional
        This is just for informational purpose and defines the source
        of the model, e.g. ABAQUS. Default is USER.

    Attributes
    ----------
    model_name : str
        Name of the main model.
    part_name : str
        Name of the part instance.
    step_name : str
        Name of the load step.
    frame_time : float
        Current frame time of the model frame.
    source : ParaqusConstant
        This is a constant for informational purpose and describes the
        source of the model, e.g. ABAQUS or USER.
    elements : ElementRepository
        A repository storing all information on elements, e.g. tags and
        element types.
    nodes : NodeRepository
        Repository storing all information on nodes, e.g. tags and
        coordinates.
    node_fields : NodeFieldRepository
        Repository storing all node fields.
    element_fields : ElementRepository
        Repository sotring all element fields.

    Example
    -------
    >>> import numpy as np
    >>> from paraqus import ParaqusModel, BinaryWriter
    >>>
    >>> # Define the model parameters
    >>> element_tags = np.array([1,2])
    >>> connectivity = [np.array([1,2,3,4]),np.array([4,3,5])]
    >>> element_types = np.array([9,5])
    >>> node_tags = np.array([1,2,3,4,5])
    >>> node_coords = np.array([[0,0],[1,0],[1,1],[0,1],[0.5,1.5]])
    >>>
    >>> # Create a new model
    >>> model = ParaqusModel(element_tags,
    >>>                      connectivity,
    >>>                      element_types,
    >>>                      node_tags,
    >>>                      node_coords,
    >>>                      model_name="foo",
    >>>                      part_name="bar",
    >>>                      step_name="foobar",
    >>>                      frame_time=0.1)
    >>>
    >>> # Export model as VTK
    >>> writer = BinaryWriter()
    >>> writer.write(model)

    """

    def __init__(self,
                 element_tags,
                 connectivity,
                 element_types,
                 node_tags,
                 node_coords,
                 **kwargs):

        # Check keyword arguments
        self.model_name = kwargs.pop("model_name", "MODEL NAME")
        self.part_name = kwargs.pop("part_name", "PART NAME")
        self.step_name = kwargs.pop("step_name", "STEP NAME")
        self.frame_time = kwargs.pop("frame_time", 0.0)
        self.source = kwargs.pop("source", USER)

        # Check for unrecognized kwargs
        if len(kwargs) > 0:
            err_msg = "Unrecognized kwarg: '{}'.".format(str(kwargs.keys()[0]))
            raise ValueError(err_msg)

        # Create model geometry
        self.elements = ElementRepository(element_tags, connectivity,
                                          element_types)
        self.nodes = NodeRepository(node_tags, node_coords)

        # Add field repositories
        self.node_fields = NodeFieldRepository()
        self.element_fields = ElementFieldRepository()

    # Methods
    def __str__(self):
        """String representation of the model."""
        description = ("ParaqusModel\n"
                       + "  model_name: '" + self.model_name + "'\n"
                       + "  part_name:  '" + self.part_name + "'\n"
                       + "  step_name:  '" + self.step_name + "'\n")

        return description

    def add_field(self, field_name, field_tags, field_values,
                         field_position, field_type):
        """
        Add a field to the model.

        The field will also be part of any exported VTK model.

        Parameters
        ----------
        field_name : str
            Name of the field.
        field_tags : numpy.ndarray
            Tags of the nodes or elements at which the field is stored.
        field_values : Sequence
            Values of the field in order of the field tags. At nodes or
            elements that are not part of the field tags, the values will be
            set to NaN automatically.
        field_position : ParaqusConstant
            A constant defining the storage position of the field, i.e.
            NODES or ELEMENTS.
        field_type : ParaqusConstant
            A constant defining the field type, i.e. SCALAR, VECTOR or
            TENSOR.

        Returns
        -------
        None.

        """
        field_values = np.asarray(field_values, dtype=float)

        # Create field object
        # An error will be thrown in case of invalid choices
        # of field_position and field_type
        field = Field(field_name, field_values,
                      field_position, field_type)

        # Add a node field
        if field_position == NODES:

            if len(self.nodes.tags) < len(field_tags):
                msg = ("Values array '{}'".format(field_name)
                       + " has more entries than the model has nodes.")
                raise ValueError(msg)

            # If there are more nodes than field values set the missing
            # value to nan
            # elif len(self.nodes.tags) > len(field_tags):
            # Always call pad method so that values will be sorted
            # regarding the order of node tags
            field._field_values = self._pad_field_values(field_tags,
                                                         field_values,
                                                         self.nodes.tags)

            self.node_fields.add_field(field)

        # Add an element field
        elif field_position == ELEMENTS:

            if len(self.elements.tags) < len(field_tags):
                msg = ("Values array '{}'".format(field_name)
                       + " has more entries than the model has elements.")
                raise ValueError(msg)

            # If there are more elements than field values set the
            # missing value to nan
            # elif len(self.elements.tags) > len(field_tags):
            # Always call pad method so that values will be sorted
            # regarding the order of element tags
            field._field_values = self._pad_field_values(field_tags,
                                                         field_values,
                                                         self.elements.tags)

            self.element_fields.add_field(field)

        else:
            msg = "Invalid field position: {}.".format(field_position)
            raise ValueError(msg)

    def add_node_group(self, group_name, node_tags):
        """
        Add a node group to the model.

        The nodes must have been added beforehand, this method just
        stores the information that they form a named group.

        Parameters
        ----------
        group_name : str
            Name of the group. Must be unique.
        node_tags : Sequence
            Integer tags of the nodes in the group.

        Returns
        -------
        None.

        """
        self.nodes.add_group(group_name, node_tags)

    def add_element_group(self, group_name, element_tags):
        """
        Add an element group to the model.

        The elements must have been added beforehand, this method just
        stores the information that the elements form a named group.

        Parameters
        ----------
        group_name : str
            Name of the group. Must be unique.
        element_tags : Sequence
            Integer tags of the elements in the group.

        Returns
        -------
        None.

        """
        self.elements.add_group(group_name, element_tags)

    def split_model(self, number_of_pieces):
        """
        Split the model into number_of_pieces parts.

        The split is performed based on element numbers, so that the
        resulting model pieces are not necessarily continuous (i.e. they
        might have holes etc).

        Parameters
        ----------
        number_of_pieces : int
            Number of parts the model will be split into.

        Yields
        ------
        piece : ParaqusModel
            The resulting submodel for one piece including all fields and
            groups.

        """
        # Split list of element tags regarding the number of pieces
        element_tags_per_piece = np.array_split(self.elements.tags,
                                                number_of_pieces)

        for piece_element_tags in element_tags_per_piece:
            # Create the piece
            if number_of_pieces > 1:
                piece = self._extract_submodel_by_elements(piece_element_tags)
            else:
                piece = self

            yield piece

    def _extract_submodel_by_elements(self, element_tags):
        """
        Extract a submodel based on element tags.

        Parameters
        ----------
        element_tags : Sequence
            List of tags of the elements that will be part of the
            submodel.

        Returns
        -------
        submodel : ParaqusModel
            The resulting submodel including all fields and groups.

        """
        # Get nodes and elements of the current piece
        sub_element_repo = self.elements.get_subset(*element_tags)
        node_tags = sub_element_repo.get_node_tags()
        sub_node_repo = self.nodes.get_subset(*node_tags)

        # Create a new paraqus model representing the submodel
        submodel = ParaqusModel(sub_element_repo.tags,
                                sub_element_repo.connectivity,
                                sub_element_repo.types,
                                sub_node_repo.tags,
                                sub_node_repo.coordinates,
                                model_name=self.model_name,
                                part_name=self.part_name,
                                step_name=self.step_name,
                                frame_time=self.frame_time,
                                source=self.source)

        # Add field to the submodel
        for nf in self.node_fields.get_subset(sub_node_repo):
            submodel.node_fields.add_field(nf)

        for ef in self.element_fields.get_subset(sub_element_repo):
            submodel.element_fields.add_field(ef)

        # Add groups to the submodel
        for group_name, group_elements in sub_element_repo.groups.items():
            submodel.add_element_group(group_name, group_elements)

        for group_name, group_nodes in sub_node_repo.groups.items():
            submodel.add_node_group(group_name, group_nodes)

        return submodel

    def get_fields_by_type(self, field_type, field_position):
        """
        Return all fields of a specific type, e.g. of type SCALAR.

        Fields stored at nodes as well as fields stored at elements can be
        returned.

        Parameters
        ----------
        field_type : ParaqusConstant
            Type of the field, i.e. TENSOR, VECTOR or SCALAR.
        field_position : ParaqusConstant
            Position of the field, i.e. NODES or ELEMENTS.

        Returns
        -------
        list of Field
            All fields that match the requested type and position.

        """
        if field_position == NODES:
            return self.node_fields.get_fields_by_type(field_type)

        elif field_position == ELEMENTS:
            return self.element_fields.get_fields_by_type(field_type)

        else:
            msg = "Invalid field position: {}".format(field_position)
            raise Exception(msg)

    def get_node_field(self, field_name):
        """
        Request a node field of a given name.

        If there is no field found, None will be returned.

        Parameters
        ----------
        field_name : str
            Name of the requested node field.

        Returns
        -------
        Field
            The requested field.

        """
        return self.node_fields[field_name]

    def get_element_field(self, field_name):
        """
        Request an element field of a given name.

        If there is no field found, None will be returned.

        Parameters
        ----------
        field_name : str
            Name of the requested element field.

        Returns
        -------
        Field
            The requested field.

        """
        return self.element_fields[field_name]

    def _pad_field_values(self, field_tags, field_values, pad_tags):
        """Extend a field value array with nan values for unset tags."""
        # TODO: force sorting early to remove the need for index juggling
        field_values = field_values.reshape(len(field_values), -1)

        # make sure the padded tags contain all of the field tags
        if not all(np.in1d(field_tags, pad_tags)):
            raise ValueError("Cannot pad values: not all tags are found.")

        # Indices of the field_tags in pad_tags
        sorter = np.argsort(pad_tags)
        indices_sorted = sorter[np.searchsorted(pad_tags, field_tags)]
        indices_original = sorter[indices_sorted]

        # New array pre-filled with nans
        shape = (len(pad_tags), field_values.shape[1])

        new_field_values = np.empty(shape, dtype=field_values.dtype)
        new_field_values[:,:] = np.nan

        # Fill in the old values
        new_field_values[indices_original,:] = field_values

        return new_field_values


class ElementRepository(object):
    """
    Repository to store a set of elements.

    Parameters
    ----------
    element_tags : Sequence
        Tags of all elements that will be stored in the repository.
    connectivity : Sequence of Sequence
        Connectivity of all elements in order of the element tags.
    element_types : Sequence
        Type of all elements in order of the element tags.

    Attributes
    ----------
    tags : numpy.ndarray
        Tags of all elements stored in the repository.
    connectivity : list of numpy.ndarray
        Connectivity list of all elements in the same order as the tags.
    types : numpy.ndarray
        Types of all elements in the same order as the tags.
    index_mapper : dict
        Mapping element tag -> index in list of tags.
    groups : dict
        Mapping group name -> element tags.

    Methods
    -------
    add_group
        Add an element group to the repository.
    get_subset
        Create a new ElementRepository containing only a subset of
        elements based on element tags.
    get_subset_by_type
        Create a new ElementRepository containing only a subset of
        elements based on element types.
    get_node_tags
        Get the tags of the nodes needed for all elements in the
        repository.

    """

    def __init__(self,
                 element_tags,
                 connectivity,
                 element_types):

        assert len(element_tags) == len(connectivity) == len(element_types)

        # The connectivity cannot be stored as one array in case of
        # elements with different numbers of element nodes
        self._tags = np.asarray(element_tags).reshape(-1)
        self._index_mapper = dict(zip(element_tags, range(len(element_tags))))
        self._connectivity = [np.asarray(c).reshape(-1) for c in connectivity]
        self._types = np.asarray(element_types, dtype=np.uint8).reshape(-1)
        self._groups = {}


    # Properties
    @property
    def tags(self):
        return self._tags

    @property
    def index_mapper(self):
        return self._index_mapper

    @property
    def connectivity(self):
        return self._connectivity

    @property
    def types(self):
        return self._types

    @property
    def groups(self):
        return self._groups


    # Methods
    def __getitem__(self, key):
        """Index-based access."""
        if key > len(self) - 1:
            raise IndexError("Index out of bounds: {}.".format(key))

        return (self.tags[key], self.connectivity[key],
                self.types[key], self.index_mapper[self.tags[key]])

    def __iter__(self):
        """Repository iterator."""
        for i in range(len(self)):
            yield self[i]

    def __len__(self):
        """Repository size."""
        return len(self.tags)

    def add_group(self, group_name, element_tags):
        """
        Add an element group to the repository.

        The elements must have been added beforehand, this method just
        stores the information that the elements form a named group.

        Parameters
        ----------
        group_name : str
            Name of the group. Must be unique.
        element_tags : Sequence
            Integer tags of the elements in the group.

        Returns
        -------
        None.

        """
        assert group_name not in self.groups, \
            "Group '{}' already exists.".format(group_name)

        assert set(element_tags) <= set(self.tags), \
            "Some element tags in the set are not in the model."

        element_tags = np.asarray(element_tags)
        element_tags.sort()
        self._groups[group_name] = element_tags


    def get_subset(self, *element_tags):
        """
        Create a new ElementRepository containing only a subset of elements.

        The subset contains all elements with the requested tags.

        Parameters
        ----------
        *element_tags : Sequence
            Elements that will be contained in the new repository.

        Returns
        -------
        ElementRepository
            Repository containing the elements of the subset.

        """
        # If there is in anvalid element tag, an error will be raised
        # by the index mapper, thus there is no need to check for valid
        # inputs here
        indices = np.asarray([self.index_mapper[e]
                              for e in element_tags])

        tags = self.tags[indices]
        connectivity = [self.connectivity[i] for i in indices]
        types = self.types[indices]

        # Generate a new element repository to return
        return_repo = ElementRepository(tags, connectivity, types)

        # Make sure the indices are coherent with the original model
        return_repo._index_mapper = dict(zip(tags, indices))

        # Extract group elements that are present in the subset
        for group_name, group_tags in self.groups.items():
            new_group_tags = np.intersect1d(group_tags, element_tags)
            return_repo.add_group(group_name, new_group_tags)

        return return_repo

    def get_subset_by_type(self, *element_types):
        """
        Create a new ElementRepository containing only a subset of elements.

        The subset contains all elements of the requested VTK types.

        Parameters
        ----------
        *element_types : Sequence
            Types of the elements that will be part of the repository.
            The element types must follow the VTK type convention.

        Returns
        -------
        ElementRepository
            Repository containing the elements of the subset.

        """
        indices = []
        for element_type in element_types:
            indices.append(np.asarray(self.types == element_type)
                           .nonzero()[0])

        # Flatten the list of indices
        indices = np.concatenate(indices)

        # Return None in case nothing was found
        if len(indices) == 0:
            return None

        # Build the new repo
        tags = self.tags[indices]
        connectivity = [self.connectivity[i] for i in indices]
        types = self.types[indices]

        return_repo = ElementRepository(tags, connectivity, types)
        return_repo.index_mapper = dict(zip(tags, indices))

        return return_repo

    def get_node_tags(self):
        """
        Return an array of node tags used for the element in the repository.

        Returns
        -------
        tags : numpy.ndarray
            Sorted node tags.

        """
        tags = np.unique(list(itertools.chain(*self.connectivity)))
        tags.sort()
        return tags


class NodeRepository(object):
    """
    Repository to store a set of nodes.

    Parameters
    ----------
    node_tags : Sequence
        Tags of all nodes that will be stored in the repository.
    node_coords : Sequence of Sequence
        Nodal coordinates in order of the node tags. The coordinates
        can be defined in 1d-, 2d- or 3d-space as an array of shape
        (number of nodes x number of dimensions).

    Attributes
    ----------
    tags : numpy.ndarray
        Tags of all nodes stored in the repository.
    coordinates : numpy.ndarray
        Coordinates of all nodes in order of the node tags.
    index_mapper : dict
        Mapping node tag -> index in list of tags.
    groups : dict
        Mapping group name -> node tags.


    """

    def __init__(self,
                 node_tags,
                 node_coords):

        assert len(node_tags) == len(node_coords)

        self._tags = np.asarray(node_tags).reshape(-1)
        self._index_mapper = dict(zip(node_tags, range(len(node_tags))))
        self._groups = {}

        node_coords = np.asarray(node_coords).reshape((len(self._tags), -1))
        rows, columns = node_coords.shape
        if columns not in (1, 2, 3):
            msg = "Invalid nodal coordinates."
            raise ValueError(msg)

        self._coordinates = node_coords

    # Properties
    @property
    def tags(self):
        return self._tags

    @property
    def index_mapper(self):
        return self._index_mapper

    @property
    def coordinates(self):
        return self._coordinates

    @property
    def groups(self):
        return self._groups


    # Methods
    def __getitem__(self, key):
        """Index-based access."""
        return (self.tags[key], self.coordinates[key],
                self.index_mapper[self.tags[key]])

    def __iter__(self):
        """Repository iterator."""
        for i in range(len(self)):
            yield self[i]

    def __len__(self):
        """Repository size."""
        return len(self.tags)

    def add_group(self, group_name, node_tags):
        """
        Add a node group to the repository.

        The nodes must have been added beforehand, this method just
        stores the information that they form a named group.

        Parameters
        ----------
        group_name : str
            Name of the group. Must be unique.
        node_tags : Sequence
            Integer tags of the nodes in the group.

        Returns
        -------
        None.

        """
        assert group_name not in self.groups, \
            "Group '{}' already exists.".format(group_name)

        assert set(node_tags) <= set(self.tags), \
            "Some node tags in the set are not in the model."

        node_tags = np.asarray(node_tags)
        node_tags.sort()
        self._groups[group_name] = node_tags

    def get_subset(self, *node_tags):
        """
        Create a new NodeRepository containing only a subset of nodes.

        The subset contains all nodes with the requested tags.

        Parameters
        ----------
        *node_tags : Sequence
            Nodes that will be contained in the new repository.

        Returns
        -------
        NodeRepository
            Repository containing the nodes of the subset.

        """
        # If there is in anvalid element tag, an error will be raised
        # by the index mapper, thus there is no need to check for valid
        # inputs here
        indices = np.array([self.index_mapper[n] for n in node_tags])

        tags = self.tags[indices]
        coords = self.coordinates[indices]

        # Generate a new element repository to return
        return_repo = NodeRepository(tags, coords)

        # make sure the indices are coherent with the original model
        return_repo._index_mapper = dict(zip(tags, indices))

        # extract group elements that are present in the subset
        for group_name, group_tags in self.groups.items():
            new_group_tags = np.intersect1d(group_tags, node_tags)
            return_repo.add_group(group_name, new_group_tags)

        return return_repo


class FieldRepositoryBaseClass(object):
    """
    Base class for field repositories.

    Attributes
    ----------
    fields : dict
        Mapping field name -> Field.

    """
    __metaclass__ = ABCMeta

    def __init__(self):
        self._fields = {}


    # Properties
    @property
    def fields(self):
        return self._fields


    # Methods
    def __getitem__(self, field_name):
        """Key-based access."""
        return self.fields[field_name]

    def __iter__(self):
        """Repository iterator."""
        return self.fields.values().__iter__()

    def __len__(self):
        """Repository size."""
        return len(self.fields)

    def get_fields_by_type(self, field_type):
        """
        Export fields based on a requested field type.

        Parameters
        ----------
        field_type : ParaqusConstant
            Type of the field, i.e. TENSOR, VECTOR or SCALAR.

        Returns
        -------
        list of Field
            Fields that match the requested type.

        """
        return [f for f in self.fields.values() if f.field_type == field_type]

    @abstractmethod
    def add_field(self, field):
        """
        Add a new node or element field to the repository.

        This method is abstract since one must differ between node and
        element fields when adding them to the respective repository.
        """
        pass

    @abstractmethod
    def get_subset(self):
        """
        Export node or element fields for a subset of nodes or elements.

        This method is abstract since one must differ between working
        with subsets of nodes or elements.
        """
        return


class NodeFieldRepository(FieldRepositoryBaseClass):
    """
    Repository to stored nodal fields.

    Attributes
    ----------
    fields : dict
        Mapping field name -> Field.


    """

    def __init__(self):
        super(NodeFieldRepository, self).__init__()

    def add_field(self, field):
        """
        Add a new node field to the repository.

        Parameters
        ----------
        field : Field
            The field with position NODES to add to the
            repository.

        Returns
        -------
        None.

        """
        if field.field_position == NODES:

            if self.fields.get(field.field_name) != None:
                msg = ("Node field '{}'".format(field.field_name)
                       + " has already been stored and will be overwritten.")
                warnings.warn(msg)

            self.fields[field.field_name] = field

        else:
            msg = ("Provided field's position is '{}' and not '{}'."
                   .format(field.field_position, NODES))
            raise ValueError(msg)

    def get_subset(self, node_repository):
        """
        Export node fields for a subset of nodes.

        The nodes are specified by a NodeRepository, which should be
        derived from a common main model by model.nodes.get_subset()
        to ensure coherent indexing.

        Parameters
        ----------
        node_repository : NodeRepository
            Node repository containing the subset of nodes.

        Returns
        -------
        sub_fields : list of Field
            The fields matching the subset of nodes.

        """
        tags = node_repository.tags
        index_mapper = node_repository.index_mapper
        indices = np.array([index_mapper[i] for i in tags])

        sub_fields = [Field(f.field_name,
                                  f.field_values[indices],
                                  f.field_position,
                                  f.field_type) for f in self.fields.values()]

        return sub_fields


class ElementFieldRepository(FieldRepositoryBaseClass):
    """
    Repository to stored elemental fields.

    Attributes
    ----------
    fields : dict
        Mapping field name -> Field.


    """

    def __init__(self):
        super(ElementFieldRepository, self).__init__()

    def add_field(self, field):
        """
        Add a new element field to the repository.

        Parameters
        ----------
        field : Field
            The field with position ELEMENTS to add to the
            repository.

        Returns
        -------
        None.

        """
        if field.field_position == ELEMENTS:

            if self.fields.get(field.field_name, None) != None:
                msg = ("Element field '{}'".format(field.field_name)
                       + " has already been stored and will be overwritten.")
                warnings.warn(msg)

            self.fields[field.field_name] = field

        else:
            msg = ("Provided field's position is '{}' and not '{}'."
                   .format(field.field_position, ELEMENTS))
            raise ValueError(msg)

    def get_subset(self, element_repository):
        """
        Export element fields for a subset of elements.

        The elements are specified by an ElementRepository, which should
        be derived from a common main model by model.elements.get_subset()
        to ensure coherent indexing.

        Parameters
        ----------
        element_repository : ElementRepository
            Element repository containing the subset of elements.

        Returns
        -------
        sub_fields : list of Field
            The fields matching the subset of elements.

        """
        tags = element_repository.tags
        index_mapper = element_repository.index_mapper
        indices = np.array([index_mapper[i] for i in tags])

        sub_fields = [Field(f.field_name,
                                  f.field_values[indices],
                                  f.field_position,
                                  f.field_type) for f in self.fields.values()]

        return sub_fields


class Field(object):
    """
    Definition of nodal or elemental fields.

    Parameters
    ----------
    field_name : str
        Name of the field.
    field_values : Sequence
        Values of the field. In case the field will be added to a model, the
        values must be in order of the respective node or element tags.
    field_position : ParaqusConstant
        Position of the field, i.e. NODES or ELEMENTS.
    field_type : ParaqusConstant
        Type of the field, i.e. TENSOR, VECTOR or SCALAR.

    Attributes
    ----------
    field_name : str
        Name of the field.
    field_position : ParaqusConstant
        Position of the field, i.e. NODES or ELEMENTS.
    field_type : ParaqusConstant
        Type of the field, i.e. TENSOR, VECTOR or SCALAR.
    field_values : numpy.ndarray
        Field values of the field.


    """

    def __init__(self,
                 field_name,
                 field_values,
                 field_position,
                 field_type):

        self.field_name = field_name

        # Add position
        if field_position not in (NODES, ELEMENTS):
            msg = "Invalid field position: {}".format(field_position)
            raise ValueError(msg)
        self._field_position = getattr(constants, str(field_position).upper())

        # Add type
        if field_type not in (SCALAR, VECTOR, TENSOR):
            msg = "Invalid field type: {}".format(field_type)
            raise ValueError(msg)
        self._field_type = getattr(constants, str(field_type).upper())

        # Add values
        field_values = np.asarray(field_values)
        if self.field_type == SCALAR:
            if np.squeeze(field_values).ndim > 1:
                msg = "Data for scalar field is not 1d."
                raise ValueError(msg)
            self._field_values = field_values.reshape((-1,1))
        else:
            self._field_values = field_values


    # Properties
    @property
    def field_values(self):
        return self._field_values

    @property
    def field_position(self):
        return self._field_position

    @property
    def field_type(self):
        return self._field_type


    # Methods
    def __repr__(self):
        descr = "Field '{}' of type '{}' at position '{}'".format(
            self.field_name, self.field_type, self.field_position)
        return descr

    def __len__(self):
        return len(self.field_values)

    def get_3d_field_values(self):
        """
        Get a copy of the field values in 3d representation.

        Vectors are in order (x, y, z) and tensors are in order
        (xx, yy, zz, xy, yz, xz). Non-symmetric tensors are not
        suppurted.

        Returns
        -------
        numpy.ndarray
            The 3d representation of the field values.

        """
        if self.field_type == SCALAR:
            return self.field_values

        elif self.field_type == VECTOR:
            nvals, ndim = self.field_values.shape
            if ndim < 3:
                return np.hstack((self.field_values, np.zeros((nvals, 3-ndim))))
            return self.field_values

        elif self.field_type == TENSOR:
            nvals, ndim = self.field_values.shape
            if ndim < 6:
                return np.hstack((self.field_values, np.zeros((nvals, 6-ndim))))
            return self.field_values

        else:
            raise NotImplementedError(
                "Only SCALAR, VECTOR and TENSOR are supported."
                )



"""
Tests for the BinaryWriter, AsciiWriter, CollectionWriter classes.

These tests can be executed in Abaqus python or in standard python 2.7+.

"""

import unittest
import numpy as np
import os
import datetime
import shutil

from paraqus.paraqusmodel import ParaqusModel
from paraqus.writers import AsciiWriter, BinaryWriter, CollectionWriter


class TestWritersGeneration(unittest.TestCase):
    
    def test_ascii_writer_generation(self):
        """AsciiWriter is initialized with correct attributes."""
        ascii_writer = AsciiWriter("test_path", 
                                   clear_output_dir=False, 
                                   number_of_pieces=2)
        
        assert ascii_writer.output_dir == os.path.abspath("test_path")
        assert ascii_writer.FORMAT == "ASCII"
        
    def test_binary_writer_generation(self):
        """BinaryWriter is initialized with correct attributes."""
        binary_writer = BinaryWriter("test_path", 
                                     clear_output_dir=False, 
                                     number_of_pieces=2)
        
        assert binary_writer.output_dir == os.path.abspath("test_path")
        assert binary_writer.FORMAT == "BINARY"
        
    def test_collection_writer_generation(self):
        """CollectionWriter is initialized with correct attributes."""
        binary_writer = BinaryWriter()
        collection_writer = CollectionWriter(binary_writer,
                                             "test_name")
        
        assert collection_writer.writer is binary_writer
        assert collection_writer.collection_name == "test_name"


class TestAsciiWriter(unittest.TestCase):
    
    def setUp(self):
        """Create a small ParaqusModel."""
        # Define geometry consisting of two first-order quad elements and
        # three first-order tri elements
        node_tags = [1, 2, 3, 4, 5, 6, 7, 8]
        node_coords = [[0, 0],
                       [1, 0],
                       [2, 0],
                       [0, 1],
                       [1, 1],
                       [2, 1],
                       [0.5, 1.5],
                       [1.5, 1.5]]
        element_tags = [1, 2, 3, 4, 5]
        connectivity = [[1, 2, 5, 4],
                        [2, 3, 6, 5],
                        [4, 5, 7],
                        [5, 6, 8],
                        [5, 8, 7]]
        element_types = [9, 9, 5, 5, 5]
    
        # Create four element base model
        model_name = "2D_TEST_MODEL"
        part_name = "2D_TEST_PART"
        self.model = ParaqusModel(element_tags,
                                  connectivity,
                                  element_types,
                                  node_tags,
                                  node_coords,
                                  model_name=model_name,
                                  part_name=part_name)
        
        self.ascii_folder = str(datetime.datetime.now()).replace(":", ".")
        
    def tearDown(self):
        """Delete dummy folder after test has finished."""
        if os.path.isdir(self.ascii_folder):
            shutil.rmtree(self.ascii_folder, ignore_errors=True)
        
    def test_write_vtu_file(self):
        """An ascii .vtu file can be written based on a ParaqusModel."""
        # Initialize writer
        ascii_writer = AsciiWriter(self.ascii_folder, 
                                   clear_output_dir=True, 
                                   number_of_pieces=1)
        
        # Export model
        ascii_writer.write(self.model)
        vtu_file = os.path.join(self.ascii_folder, "2D_TEST_MODEL", 
                                "vtu", "2D_TEST_PART_0_0.vtu")
        
        assert os.path.isfile(vtu_file)
        
    def test_write_pvtu_file(self):
        """Multiple ascii .vtu files can be written based on a ParaqusModel."""
        # Initialize writer
        ascii_writer = AsciiWriter(self.ascii_folder, 
                                   clear_output_dir=True, 
                                   number_of_pieces=2)
        
        # Export model
        ascii_writer.write(self.model)
        vtu_file_1 = os.path.join(self.ascii_folder, "2D_TEST_MODEL",
                                  "vtu", "2D_TEST_PART_0_0.vtu")
        vtu_file_2 = os.path.join(self.ascii_folder, "2D_TEST_MODEL",
                                  "vtu", "2D_TEST_PART_1_0.vtu")
        pvtu_file = os.path.join(self.ascii_folder, "2D_TEST_MODEL",
                                 "vtu", "2D_TEST_PART_0.pvtu")
        
        assert os.path.isfile(vtu_file_1)
        assert os.path.isfile(vtu_file_2)
        assert os.path.isfile(pvtu_file)
        

class TestBinaryWriter(unittest.TestCase):
    
    def setUp(self):
        """Create a small ParaqusModel."""
        # Define geometry consisting of two first-order quad elements and
        # three first-order tri elements
        node_tags = [1, 2, 3, 4, 5, 6, 7, 8]
        node_coords = [[0, 0],
                       [1, 0],
                       [2, 0],
                       [0, 1],
                       [1, 1],
                       [2, 1],
                       [0.5, 1.5],
                       [1.5, 1.5]]
        element_tags = [1, 2, 3, 4, 5]
        connectivity = [[1, 2, 5, 4],
                        [2, 3, 6, 5],
                        [4, 5, 7],
                        [5, 6, 8],
                        [5, 8, 7]]
        element_types = [9, 9, 5, 5, 5]
    
        # Create four element base model
        model_name = "2D_TEST_MODEL"
        part_name = "2D_TEST_PART"
        self.model = ParaqusModel(element_tags,
                                  connectivity,
                                  element_types,
                                  node_tags,
                                  node_coords,
                                  model_name=model_name,
                                  part_name=part_name)
        
        self.binary_folder = str(datetime.datetime.now()).replace(":", ".")
        
    def tearDown(self):
        """Delete dummy folder after test has finished."""
        if os.path.isdir(self.binary_folder):
            shutil.rmtree(self.binary_folder, ignore_errors=True)
        
    def test_write_vtu_file_base64(self):
        """A base64 .vtu file can be written based on a ParaqusModel."""
        # Initialize writer
        binary_writer = BinaryWriter(self.binary_folder, 
                                     clear_output_dir=True, 
                                     number_of_pieces=1)
        
        # Export model
        binary_writer.write(self.model)
        vtu_file = os.path.join(self.binary_folder, "2D_TEST_MODEL", 
                                "vtu", "2D_TEST_PART_0_0.vtu")
        
        assert os.path.isfile(vtu_file)
        
    def test_write_pvtu_file_base64(self):
        """Multiple base64 .vtu files can be written based on a ParaqusModel."""
        # Initialize writer
        binary_writer = BinaryWriter(self.binary_folder, 
                                     clear_output_dir=True, 
                                     number_of_pieces=2)
        
        # Export model
        binary_writer.write(self.model)
        vtu_file_1 = os.path.join(self.binary_folder, "2D_TEST_MODEL",
                                  "vtu", "2D_TEST_PART_0_0.vtu")
        vtu_file_2 = os.path.join(self.binary_folder, "2D_TEST_MODEL",
                                  "vtu", "2D_TEST_PART_1_0.vtu")
        pvtu_file = os.path.join(self.binary_folder, "2D_TEST_MODEL",
                                 "vtu", "2D_TEST_PART_0.pvtu")
        
        assert os.path.isfile(vtu_file_1)
        assert os.path.isfile(vtu_file_2)
        assert os.path.isfile(pvtu_file)
        
    def test_write_vtu_file_raw(self):
        """A raw .vtu file can be written based on a ParaqusModel."""
        # Initialize writer
        binary_writer = BinaryWriter(self.binary_folder, 
                                     clear_output_dir=True, 
                                     number_of_pieces=1,
                                     encoding="raw")
        
        # Export model
        binary_writer.write(self.model)
        vtu_file = os.path.join(self.binary_folder, "2D_TEST_MODEL", 
                                "vtu", "2D_TEST_PART_0_0.vtu")
        
        assert os.path.isfile(vtu_file)
        
    def test_write_pvtu_file_raw(self):
        """Multiple raw .vtu files can be written based on a ParaqusModel."""
        # Initialize writer
        binary_writer = BinaryWriter(self.binary_folder, 
                                     clear_output_dir=True, 
                                     number_of_pieces=2,
                                     encoding="raw")
        
        # Export model
        binary_writer.write(self.model)
        vtu_file_1 = os.path.join(self.binary_folder, "2D_TEST_MODEL",
                                  "vtu", "2D_TEST_PART_0_0.vtu")
        vtu_file_2 = os.path.join(self.binary_folder, "2D_TEST_MODEL",
                                  "vtu", "2D_TEST_PART_1_0.vtu")
        pvtu_file = os.path.join(self.binary_folder, "2D_TEST_MODEL",
                                 "vtu", "2D_TEST_PART_0.pvtu")
        
        assert os.path.isfile(vtu_file_1)
        assert os.path.isfile(vtu_file_2)
        assert os.path.isfile(pvtu_file)


class TestCollectionWriter(unittest.TestCase):
    
    def setUp(self):
        """Create a small ParaqusModel."""
        # Define geometry consisting of two first-order quad elements and
        # three first-order tri elements
        node_tags = [1, 2, 3, 4, 5, 6, 7, 8]
        node_coords = [[0, 0],
                       [1, 0],
                       [2, 0],
                       [0, 1],
                       [1, 1],
                       [2, 1],
                       [0.5, 1.5],
                       [1.5, 1.5]]
        element_tags = [1, 2, 3, 4, 5]
        connectivity = [[1, 2, 5, 4],
                        [2, 3, 6, 5],
                        [4, 5, 7],
                        [5, 6, 8],
                        [5, 8, 7]]
        element_types = [9, 9, 5, 5, 5]
    
        # Create four element base model
        model_name = "2D_TEST_MODEL"
        part_name = "2D_TEST_PART"
        self.model = ParaqusModel(element_tags,
                                  connectivity,
                                  element_types,
                                  node_tags,
                                  node_coords,
                                  model_name=model_name,
                                  part_name=part_name)
        
        self.collection_folder = str(datetime.datetime.now()).replace(":", ".")
        
    def tearDown(self):
        """Delete dummy folder after test has finished."""
        if os.path.isdir(self.collection_folder):
            shutil.rmtree(self.collection_folder, ignore_errors=True)
            
    def test_write_pvd_file(self):
        """A .pvd file can be written as a collection of .vtu files."""
        # Initialize writer
        binary_writer = BinaryWriter(self.collection_folder, 
                                     clear_output_dir=True, 
                                     number_of_pieces=2)
        
        # Make pvd file
        with CollectionWriter(binary_writer, "2D_TEST_MODEL") as writer:
            writer.write(self.model)
            
        vtu_file_1 = os.path.join(self.collection_folder, "2D_TEST_MODEL",
                                  "vtu", "2D_TEST_PART_0_0.vtu")
        vtu_file_2 = os.path.join(self.collection_folder, "2D_TEST_MODEL",
                                  "vtu", "2D_TEST_PART_1_0.vtu")
        pvtu_file = os.path.join(self.collection_folder, "2D_TEST_MODEL",
                                 "vtu", "2D_TEST_PART_0.pvtu")
        pvd_file = os.path.join(self.collection_folder, "2D_TEST_MODEL",
                                "2D_TEST_MODEL.pvd")
        
        assert os.path.isfile(vtu_file_1)
        assert os.path.isfile(vtu_file_2)
        assert os.path.isfile(pvtu_file)
        assert os.path.isfile(pvd_file)


# -*- coding: utf-8 -*-
#
#   Paraqus - A VTK exporter for FEM results.
#
#   Copyright (C) 2022, Furlan, Stollberg and Menzel
#
#    This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along with this program. If not, see https://www.gnu.org/licenses/.
"""
Helper functions/types to read results from abaqus odbs.

All of these can only be executed in abaqus python.

"""
import os.path
from abaqus import session

class ODBObject(object):
    """
    Context manager for Abaqus odb objects.

    Opens an odb and closes it after we are done with it. If any exceptions
    are raised while the odb is open, it is still closed afterwards.

    The optional argument `readonly` can be set to ``False`` to be able to
    write to the odb.

    """

    def __init__(self, file_name, readonly=True):
        self.file_path = os.path.abspath(file_name)
        self.readonly = readonly
        self.already_open = False

    def __enter__(self):

        # make sure to compare with absolute paths
        keys = [os.path.abspath(k) for k in session.odbs.keys()]

        # if the odb is already open, just return the odb object
        if self.file_path in keys:
            # odb is already open
            self.already_open = True
            return session.odbs[self.file_path]

        # otherwise, open it and return the object
        upgrade_odb(self.file_path)
        odb = session.openOdb(name=self.file_path, readOnly=self.readonly)

        # deal with silent errors in the readonly status, this happens e.g.
        # when a lock file prevents write access
        assert odb.isReadOnly == self.readonly, \
            "The odb could not be opened with option readonly=%s" % self.readonly

        self.odb = odb

        return odb

    def __exit__(self, type, value, traceback):
        # only close the odb if it was not open before we "opened" it
        if not self.already_open:
            self.odb.close()


def upgrade_odb(odb_file):
    """
    Upgrade an odb if necessary.

    A subfolder is created for the original odb files if an upgrade is
    performed.

    Parameters
    ----------
    odb_file : str
        Path to the Abaqus .odb

    Returns
    -------
    bool
        Whether the odb was updated.

    """
    import os
    import shutil
    from abaqus import session

    upgrade_required = session.isUpgradeRequiredForOdb(odb_file)

    if upgrade_required:
        # make sure we work with absolute paths
        odb_file = os.path.abspath(odb_file)

        # create new directory for upgraded files
        new_directory = os.path.join(os.path.dirname(odb_file),
                                    "odbs_before_upgrades")

        if not os.path.isdir(new_directory):
            os.mkdir(new_directory)

        backup_file_name = os.path.basename(odb_file)

        backup_file_path = os.path.join(new_directory, backup_file_name)

        # move the original odb to the backup folder
        shutil.move(odb_file, backup_file_path)

        # upgrade the odb file
        session.upgradeOdb(existingOdbPath=backup_file_path,
                           upgradedOdbPath=odb_file)

        print("The odb file '%s' has been updated." % odb_file)
        print("The original file has been stored in the path '%s'"
              % backup_file_path)

        return True

    else:
        return False




"""
Tests for the ODBReader class.

These tests can only be executed in Abaqus python.

"""
import os
import unittest

import numpy as np

from paraqus.abaqus import ODBReader

RESOURCE_PATH  = os.path.join(os.path.dirname(os.path.dirname(__file__)),
                              "resources")

class TestElementTestMesh(unittest.TestCase):
    """Class docsrt"""
    def setUp(self):
        self.odb_path = os.path.join(RESOURCE_PATH, 
                                     'element_test_current.odb')
        
        self.step_name = "Apply Force"
        self.frame_index = -1
        self.instance_name = "Rectangular Instance"
        
        self.reader = ODBReader(odb_path=self.odb_path,
                                    model_name='test_model')
            

    def test_read_mesh(self):
        """The model built from the element test odb has the correct mesh."""  
        
        instance_models = list(
            self.reader.read_instances(step_name = self.step_name,
                                       frame_index = self.frame_index)
                              )
    
        assert len(instance_models) == 1
            
        instance_model = instance_models[0]
        
        assert len(instance_model.nodes) == 9
        assert np.all(instance_model.nodes.tags == [1,2,3,4,5,6,7,8,9])
        
        assert len(instance_model.elements) == 4
        assert np.all(instance_model.elements.tags == [1,2,3,4])
        
    
    def test_read_node_field(self):
        """The model has a nodal vector field 'U'."""
        self.reader.add_field_export_request('U')
        
        instance_models = list(
            self.reader.read_instances(step_name = self.step_name,
                                       frame_index = self.frame_index)
                              )
    
        assert len(instance_models) == 1
            
        instance_model = instance_models[0]
        
        u = instance_model.get_node_field('U')
        assert len(u.field_values) == 9
        
        assert u.field_type == "vector"
        assert u.field_position == "nodes"
        
        
    def test_read_element_field(self):
        """The model has an elemental tensor field 'S'."""
        self.reader.add_field_export_request('S')
        
        instance_models = list(
            self.reader.read_instances(step_name = self.step_name,
                                       frame_index = self.frame_index)
                              )
    
        assert len(instance_models) == 1
            
        instance_model = instance_models[0]
        
        s = instance_model.get_element_field('S')
        assert len(s.field_values) == 4 #1 value per element
        
        assert s.field_type == "tensor"
        assert s.field_position == "elements"
        
        
    def test_error_missing_node_field(self):
        """A KeyError is raised for missing node fields."""
        instance_models = list(
            self.reader.read_instances(step_name = self.step_name,
                                       frame_index = self.frame_index)
                              )
    
        assert len(instance_models) == 1
            
        instance_model = instance_models[0]
        
        with self.assertRaises(KeyError) as context:
            s = instance_model.get_node_field('U')
         
        
    def test_error_missing_element_field(self):
        """A KeyError is raised for missing element fields."""
        instance_models = list(
            self.reader.read_instances(step_name = self.step_name,
                                       frame_index = self.frame_index)
                              )
    
        assert len(instance_models) == 1
            
        instance_model = instance_models[0]
        
        with self.assertRaises(KeyError) as context:
            s = instance_model.get_element_field('S')
            
            
    def test_read_instance_node_set(self):
        """A group is stored in the model for an instance node set."""
        self.reader.add_set_export_request('Left Edge',
                                           set_type="nodes",
                                           instance_name=self.instance_name)
        
        instance_models = list(
            self.reader.read_instances(step_name = self.step_name,
                                       frame_index = self.frame_index)
                              )
    
        assert len(instance_models) == 1
            
        instance_model = instance_models[0]
        
        # instance-level sets have the instance name prepended
        grp_name = self.instance_name + '.' + 'Left Edge'
        
        grp_nodes = instance_model.nodes.groups[grp_name]
        
        assert len(grp_nodes) == 3
    
    def test_read_instance_element_set(self):
        """A group is stored in the model for an instance element set."""
        self.reader.add_set_export_request('Left Edge',
                                           set_type="elements",
                                           instance_name=self.instance_name)
        
        instance_models = list(
            self.reader.read_instances(step_name = self.step_name,
                                       frame_index = self.frame_index)
                              )
    
        assert len(instance_models) == 1
            
        instance_model = instance_models[0]
        
        # instance-level sets have the instance name prepended
        grp_name = self.instance_name + '.' + 'Left Edge'
        
        grp_nodes = instance_model.elements.groups[grp_name]
        
        assert len(grp_nodes) == 2
        
        
    def test_read_assembly_node_set(self):
        """A group is stored in the model for an assembly node set."""
        self.reader.add_set_export_request('Bottom Edge',
                                           set_type="nodes")
        
        instance_models = list(
            self.reader.read_instances(step_name = self.step_name,
                                       frame_index = self.frame_index)
                              )
    
        assert len(instance_models) == 1
            
        instance_model = instance_models[0]
        
        # instance-level sets have the instance name prepended
        grp_name = 'Bottom Edge'
        
        grp_nodes = instance_model.nodes.groups[grp_name]
        
        assert len(grp_nodes) == 3
    
    
    def test_read_assembly_element_set(self):
        """A group is stored in the model for an assembly element set."""
        self.reader.add_set_export_request('Bottom Edge',
                                           set_type="elements")
        
        instance_models = list(
            self.reader.read_instances(step_name = self.step_name,
                                       frame_index = self.frame_index)
                              )
    
        assert len(instance_models) == 1
            
        instance_model = instance_models[0]
        
        # instance-level sets have the instance name prepended
        grp_name = 'Bottom Edge'
        
        grp_nodes = instance_model.elements.groups[grp_name]
        
        assert len(grp_nodes) == 2
        
    
    def test_read_instance_surface_nodes(self):
        """A group is stored in the model for nodes of an instance surface."""
        self.reader.add_surface_export_request('Right Edge',
                                               surface_type="nodes",
                                               instance_name=self.instance_name)
        
        instance_models = list(
            self.reader.read_instances(step_name = self.step_name,
                                       frame_index = self.frame_index)
                              )
    
        assert len(instance_models) == 1
            
        instance_model = instance_models[0]
        
        # instance-level sets have the instance name prepended
        grp_name = 'surface.%s.Right Edge' % self.instance_name
        
        grp_nodes = instance_model.nodes.groups[grp_name]
        
        assert len(grp_nodes) == 3
        
        
    def test_read_instance_surface_elements(self):
        """A group is stored in the model for elements of an instance surface."""
        self.reader.add_surface_export_request('Right Edge',
                                               surface_type="elements",
                                               instance_name=self.instance_name)
        
        instance_models = list(
            self.reader.read_instances(step_name = self.step_name,
                                       frame_index = self.frame_index)
                              )
    
        assert len(instance_models) == 1
            
        instance_model = instance_models[0]
        
        # instance-level sets have the instance name prepended
        grp_name = 'surface.%s.Right Edge' % self.instance_name
        
        grp_elements = instance_model.elements.groups[grp_name]
        
        assert len(grp_elements) == 2
    
    
    def test_read_assembly_surface_nodes(self):
        """A group is stored in the model for nodes of an assembly surface."""
        self.reader.add_surface_export_request('Top Surface',
                                               surface_type="nodes")
        
        instance_models = list(
            self.reader.read_instances(step_name = self.step_name,
                                       frame_index = self.frame_index)
                              )
    
        assert len(instance_models) == 1
            
        instance_model = instance_models[0]
        
        grp_name = 'surface.Top Surface'
        
        grp_nodes = instance_model.nodes.groups[grp_name]
        
        assert len(grp_nodes) == 3
        
        
    def test_read_assembly_surface_elements(self):
        """A group is stored in the model for elements of an assembly surface."""
        self.reader.add_surface_export_request('Top Surface',
                                               surface_type="elements")
        
        instance_models = list(
            self.reader.read_instances(step_name = self.step_name,
                                       frame_index = self.frame_index)
                              )
    
        assert len(instance_models) == 1
            
        instance_model = instance_models[0]
        
        grp_name = 'surface.Top Surface'
        
        grp_elements = instance_model.elements.groups[grp_name]
        
        assert len(grp_elements) == 2
        
        
    def test_multiple_sets(self):
        """Multiple sets and surfaces can be exported."""
        self.reader.add_set_export_request('Bottom Edge',
                                           set_type="nodes")
        
        self.reader.add_set_export_request('Bottom Edge',
                                           set_type="elements")
        
        self.reader.add_set_export_request('Left Edge',
                                           set_type="nodes",
                                           instance_name=self.instance_name)
        
        self.reader.add_set_export_request('Left Edge',
                                           set_type="elements",
                                           instance_name=self.instance_name)
        
        self.reader.add_surface_export_request('Top Surface',
                                               surface_type="nodes")
        
        self.reader.add_surface_export_request('Top Surface',
                                               surface_type="elements")
        
        self.reader.add_surface_export_request('Right Edge',
                                               surface_type="nodes",
                                               instance_name=self.instance_name)
        
        self.reader.add_surface_export_request('Right Edge',
                                               surface_type="elements",
                                               instance_name=self.instance_name)
        
        instance_models = list(
            self.reader.read_instances(step_name = self.step_name,
                                       frame_index = self.frame_index)
                              )
    
        assert len(instance_models) == 1
            
        instance_model = instance_models[0]
        
        assert len(instance_model.nodes.groups) == 4
        assert len(instance_model.elements.groups) == 4
        